@@ -132,7 +132,7 @@ func (s *Server) codeAction(ctx context.Context, params *protocol.CodeActionPara
 					codeActions = append(codeActions, protocol.CodeAction{
 						Title: importFixTitle(importFix.Fix),
 						Kind:  protocol.QuickFix,
-						Edit: protocol.WorkspaceEdit{
+						Edit: &protocol.WorkspaceEdit{
 							DocumentChanges: documentChanges(fh, importFix.Edits),
 						},
 						Diagnostics: fixes,
@@ -146,7 +146,7 @@ func (s *Server) codeAction(ctx context.Context, params *protocol.CodeActionPara
 				codeActions = append(codeActions, protocol.CodeAction{
 					Title: "Organize Imports",
 					Kind:  protocol.SourceOrganizeImports,
-					Edit: protocol.WorkspaceEdit{
+					Edit: &protocol.WorkspaceEdit{
 						DocumentChanges: documentChanges(fh, importEdits),
 					},
 				})
@@ -158,7 +158,7 @@ func (s *Server) codeAction(ctx context.Context, params *protocol.CodeActionPara
 
 		// Type-check the package and also run analysis,
 		// then combine their diagnostics.
-		pkg, _, err := source.PackageForFile(ctx, snapshot, fh.URI(), source.TypecheckFull, source.NarrowestPackage)
+		pkg, _, err := source.PackageForFile(ctx, snapshot, fh.URI(), source.NarrowestPackage)
 		if err != nil {
 			return nil, err
 		}
@@ -418,22 +418,12 @@ func codeActionsForDiagnostic(ctx context.Context, snapshot source.Snapshot, sd
 			if err != nil {
 				return nil, err
 			}
-			changes = append(changes, protocol.DocumentChanges{
-				TextDocumentEdit: &protocol.TextDocumentEdit{
-					TextDocument: protocol.OptionalVersionedTextDocumentIdentifier{
-						Version: fh.Version(),
-						TextDocumentIdentifier: protocol.TextDocumentIdentifier{
-							URI: protocol.URIFromSpanURI(fh.URI()),
-						},
-					},
-					Edits: edits,
-				},
-			})
+			changes = append(changes, documentChanges(fh, edits)...)
 		}
 		action := protocol.CodeAction{
 			Title: fix.Title,
 			Kind:  fix.ActionKind,
-			Edit: protocol.WorkspaceEdit{
+			Edit: &protocol.WorkspaceEdit{
 				DocumentChanges: changes,
 			},
 			Command: fix.Command,
