@@ -192,14 +192,14 @@ type Server struct {
 	ClientID     string
 }
 
-// AddClient adds a client to the set being served.
+// addClient adds a client to the set being served.
 func (st *State) addClient(session *cache.Session) {
 	st.mu.Lock()
 	defer st.mu.Unlock()
 	st.clients = append(st.clients, &Client{Session: session})
 }
 
-// DropClient removes a client from the set being served.
+// dropClient removes a client from the set being served.
 func (st *State) dropClient(session *cache.Session) {
 	st.mu.Lock()
 	defer st.mu.Unlock()
@@ -213,7 +213,7 @@ func (st *State) dropClient(session *cache.Session) {
 	}
 }
 
-// AddServer adds a server to the set being queried. In practice, there should
+// updateServer updates a server to the set being queried. In practice, there should
 // be at most one remote server.
 func (st *State) updateServer(server *Server) {
 	st.mu.Lock()
@@ -232,7 +232,7 @@ func (st *State) updateServer(server *Server) {
 	st.servers = append(st.servers, server)
 }
 
-// DropServer drops a server from the set being queried.
+// dropServer drops a server from the set being queried.
 func (st *State) dropServer(id string) {
 	st.mu.Lock()
 	defer st.mu.Unlock()
@@ -461,6 +461,13 @@ func (i *Instance) Serve(ctx context.Context, addr string) (string, error) {
 		mux.HandleFunc("/info", render(InfoTmpl, i.getInfo))
 		mux.HandleFunc("/memory", render(MemoryTmpl, getMemory))
 
+		// Internal debugging helpers.
+		mux.HandleFunc("/_dogc", func(w http.ResponseWriter, r *http.Request) {
+			runtime.GC()
+			runtime.GC()
+			runtime.GC()
+			http.Error(w, "OK", 200)
+		})
 		mux.HandleFunc("/_makeabug", func(w http.ResponseWriter, r *http.Request) {
 			bug.Report("bug here", nil)
 			http.Error(w, "made a bug", http.StatusOK)
@@ -540,16 +547,6 @@ func (i *Instance) writeMemoryDebug(threshold uint64, withNames bool) error {
 		return err
 	}
 
-	for _, cache := range i.State.Caches() {
-		cf, err := zipw.Create(fmt.Sprintf("cache-%v.html", cache.ID()))
-		if err != nil {
-			return err
-		}
-		if _, err := cf.Write([]byte(cache.PackageStats(withNames))); err != nil {
-			return err
-		}
-	}
-
 	if err := zipw.Close(); err != nil {
 		return err
 	}
@@ -727,7 +724,6 @@ Unknown page
 {{define "serverlink"}}<a href="/server/{{.}}">Server {{.}}</a>{{end}}
 {{define "sessionlink"}}<a href="/session/{{.}}">Session {{.}}</a>{{end}}
 {{define "viewlink"}}<a href="/view/{{.}}">View {{.}}</a>{{end}}
-{{define "filelink"}}<a href="/file/{{.Session}}/{{.FileIdentity.Hash}}">{{.FileIdentity.URI}}</a>{{end}}
 `)).Funcs(template.FuncMap{
 	"fuint64":  fuint64,
 	"fuint32":  fuint32,
@@ -764,8 +760,6 @@ var MainTmpl = template.Must(template.Must(BaseTemplate.Clone()).Parse(`
 <ul>{{range .State.Caches}}<li>{{template "cachelink" .ID}}</li>{{end}}</ul>
 <h2>Sessions</h2>
 <ul>{{range .State.Sessions}}<li>{{template "sessionlink" .ID}} from {{template "cachelink" .Cache.ID}}</li>{{end}}</ul>
-<h2>Views</h2>
-<ul>{{range .State.Views}}<li>{{.Name}} is {{template "viewlink" .ID}} from {{template "sessionlink" .Session.ID}} in {{.Folder}}</li>{{end}}</ul>
 <h2>Clients</h2>
 <ul>{{range .State.Clients}}<li>{{template "clientlink" .Session.ID}}</li>{{end}}</ul>
 <h2>Servers</h2>
@@ -824,8 +818,6 @@ var CacheTmpl = template.Must(template.Must(BaseTemplate.Clone()).Parse(`
 {{define "body"}}
 <h2>memoize.Store entries</h2>
 <ul>{{range $k,$v := .MemStats}}<li>{{$k}} - {{$v}}</li>{{end}}</ul>
-<h2>Per-package usage - not accurate, for guidance only</h2>
-{{.PackageStats true}}
 {{end}}
 `))
 
@@ -878,7 +870,11 @@ From: <b>{{template "cachelink" .Cache.ID}}</b><br>
 <h2>Views</h2>
 <ul>{{range .Views}}<li>{{.Name}} is {{template "viewlink" .ID}} in {{.Folder}}</li>{{end}}</ul>
 <h2>Overlays</h2>
-<ul>{{range .Overlays}}<li>{{template "filelink" .}}</li>{{end}}</ul>
+{{$session := .}}
+<ul>{{range .Overlays}}
+<li>
+<a href="/file/{{$session.ID}}/{{.FileIdentity.Hash}}">{{.FileIdentity.URI}}</a>
+</li>{{end}}</ul>
 <h2>Options</h2>
 {{range options .}}
 <p><b>{{.Name}}</b> {{.Type}}</p>
@@ -902,7 +898,6 @@ var FileTmpl = template.Must(template.Must(BaseTemplate.Clone()).Parse(`
 {{define "title"}}Overlay {{.FileIdentity.Hash}}{{end}}
 {{define "body"}}
 {{with .}}
-	From: <b>{{template "sessionlink" .Session}}</b><br>
 	URI: <b>{{.URI}}</b><br>
 	Identifier: <b>{{.FileIdentity.Hash}}</b><br>
 	Version: <b>{{.Version}}</b><br>
