@@ -5,7 +5,7 @@
 package fake
 
 import (
-	"bufio"
+	"bytes"
 	"context"
 	"errors"
 	"fmt"
@@ -17,6 +17,7 @@ import (
 	"sync"
 
 	"golang.org/x/tools/gopls/internal/lsp/command"
+	"golang.org/x/tools/gopls/internal/lsp/glob"
 	"golang.org/x/tools/gopls/internal/lsp/protocol"
 	"golang.org/x/tools/gopls/internal/lsp/source"
 	"golang.org/x/tools/gopls/internal/span"
@@ -38,10 +39,12 @@ type Editor struct {
 	sandbox    *Sandbox
 	defaultEnv map[string]string
 
-	mu                 sync.Mutex                  // guards config, buffers, serverCapabilities
+	// TODO(adonovan): buffers should be keyed by protocol.DocumentURI.
+	mu                 sync.Mutex
 	config             EditorConfig                // editor configuration
 	buffers            map[string]buffer           // open buffers (relative path -> buffer content)
 	serverCapabilities protocol.ServerCapabilities // capabilities / options
+	watchPatterns      []*glob.Glob                // glob patterns to watch
 
 	// Call metrics for the purpose of expectations. This is done in an ad-hoc
 	// manner for now. Perhaps in the future we should do something more
@@ -58,19 +61,14 @@ type CallCounts struct {
 
 // buffer holds information about an open buffer in the editor.
 type buffer struct {
-	windowsLineEndings bool     // use windows line endings when merging lines
-	version            int      // monotonic version; incremented on edits
-	path               string   // relative path in the workspace
-	lines              []string // line content
-	dirty              bool     // if true, content is unsaved (TODO(rfindley): rename this field)
+	version int              // monotonic version; incremented on edits
+	path    string           // relative path in the workspace
+	mapper  *protocol.Mapper // buffer content
+	dirty   bool             // if true, content is unsaved (TODO(rfindley): rename this field)
 }
 
 func (b buffer) text() string {
-	eol := "\n"
-	if b.windowsLineEndings {
-		eol = "\r\n"
-	}
-	return strings.Join(b.lines, eol)
+	return string(b.mapper.Content)
 }
 
 // EditorConfig configures the editor's LSP session. This is similar to
@@ -108,7 +106,7 @@ type EditorConfig struct {
 	Settings map[string]interface{}
 }
 
-// NewEditor Creates a new Editor.
+// NewEditor creates a new Editor.
 func NewEditor(sandbox *Sandbox, config EditorConfig) *Editor {
 	return &Editor{
 		buffers:    make(map[string]buffer),
@@ -125,14 +123,14 @@ func NewEditor(sandbox *Sandbox, config EditorConfig) *Editor {
 // It returns the editor, so that it may be called as follows:
 //
 //	editor, err := NewEditor(s).Connect(ctx, conn, hooks)
-func (e *Editor) Connect(ctx context.Context, connector servertest.Connector, hooks ClientHooks) (*Editor, error) {
+func (e *Editor) Connect(ctx context.Context, connector servertest.Connector, hooks ClientHooks, skipApplyEdits bool) (*Editor, error) {
 	bgCtx, cancelConn := context.WithCancel(xcontext.Detach(ctx))
 	conn := connector.Connect(bgCtx)
 	e.cancelConn = cancelConn
 
 	e.serverConn = conn
 	e.Server = protocol.ServerDispatcher(conn)
-	e.client = &Client{editor: e, hooks: hooks}
+	e.client = &Client{editor: e, hooks: hooks, skipApplyEdits: skipApplyEdits}
 	conn.Go(bgCtx,
 		protocol.Handlers(
 			protocol.ClientHandler(e.client,
@@ -244,6 +242,7 @@ func (e *Editor) settingsLocked() map[string]interface{} {
 
 func (e *Editor) initialize(ctx context.Context) error {
 	params := &protocol.ParamInitialize{}
+	params.ClientInfo = &protocol.Msg_XInitializeParams_clientInfo{}
 	params.ClientInfo.Name = "fakeclient"
 	params.ClientInfo.Version = "v1.0.0"
 	e.mu.Lock()
@@ -257,7 +256,7 @@ func (e *Editor) initialize(ctx context.Context) error {
 	params.Capabilities.TextDocument.Completion.CompletionItem.TagSupport.ValueSet = []protocol.CompletionItemTag{protocol.ComplDeprecated}
 
 	params.Capabilities.TextDocument.Completion.CompletionItem.SnippetSupport = true
-	params.Capabilities.TextDocument.SemanticTokens.Requests.Full = true
+	params.Capabilities.TextDocument.SemanticTokens.Requests.Full.Value = true
 	// copied from lsp/semantic.go to avoid import cycle in tests
 	params.Capabilities.TextDocument.SemanticTokens.TokenTypes = []string{
 		"namespace", "type", "class", "enum", "interface",
@@ -281,7 +280,8 @@ func (e *Editor) initialize(ctx context.Context) error {
 		},
 	}
 
-	params.Trace = "messages"
+	trace := protocol.TraceValues("messages")
+	params.Trace = &trace
 	// TODO: support workspace folders.
 	if e.Server != nil {
 		resp, err := e.Server.Initialize(ctx, params)
@@ -323,7 +323,7 @@ func (e *Editor) makeWorkspaceFoldersLocked() (folders []protocol.WorkspaceFolde
 
 // onFileChanges is registered to be called by the Workdir on any writes that
 // go through the Workdir API. It is called synchronously by the Workdir.
-func (e *Editor) onFileChanges(ctx context.Context, evts []FileEvent) {
+func (e *Editor) onFileChanges(ctx context.Context, evts []protocol.FileEvent) {
 	if e.Server == nil {
 		return
 	}
@@ -339,32 +339,42 @@ func (e *Editor) onFileChanges(ctx context.Context, evts []FileEvent) {
 	go func() {
 		e.mu.Lock()
 		defer e.mu.Unlock()
-		var lspevts []protocol.FileEvent
 		for _, evt := range evts {
 			// Always send an on-disk change, even for events that seem useless
 			// because they're shadowed by an open buffer.
-			lspevts = append(lspevts, evt.ProtocolEvent)
-
-			if buf, ok := e.buffers[evt.Path]; ok {
+			path := e.sandbox.Workdir.URIToPath(evt.URI)
+			if buf, ok := e.buffers[path]; ok {
 				// Following VS Code, don't honor deletions or changes to dirty buffers.
-				if buf.dirty || evt.ProtocolEvent.Type == protocol.Deleted {
+				if buf.dirty || evt.Type == protocol.Deleted {
 					continue
 				}
 
-				content, err := e.sandbox.Workdir.ReadFile(evt.Path)
+				content, err := e.sandbox.Workdir.ReadFile(path)
 				if err != nil {
 					continue // A race with some other operation.
 				}
 				// No need to update if the buffer content hasn't changed.
-				if content == buf.text() {
+				if string(content) == buf.text() {
 					continue
 				}
 				// During shutdown, this call will fail. Ignore the error.
-				_ = e.setBufferContentLocked(ctx, evt.Path, false, lines(content), nil)
+				_ = e.setBufferContentLocked(ctx, path, false, content, nil)
 			}
 		}
+		var matchedEvts []protocol.FileEvent
+		for _, evt := range evts {
+			filename := filepath.ToSlash(evt.URI.SpanURI().Filename())
+			for _, g := range e.watchPatterns {
+				if g.Match(filename) {
+					matchedEvts = append(matchedEvts, evt)
+					break
+				}
+			}
+		}
+
+		// TODO(rfindley): don't send notifications while locked.
 		e.Server.DidChangeWatchedFiles(ctx, &protocol.DidChangeWatchedFilesParams{
-			Changes: lspevts,
+			Changes: matchedEvts,
 		})
 	}()
 }
@@ -380,16 +390,43 @@ func (e *Editor) OpenFile(ctx context.Context, path string) error {
 	if err != nil {
 		return err
 	}
+	if e.Config().WindowsLineEndings {
+		content = toWindowsLineEndings(content)
+	}
 	return e.createBuffer(ctx, path, false, content)
 }
 
+// toWindowsLineEndings checks whether content has windows line endings.
+//
+// If so, it returns content unmodified. If not, it returns a new byte slice modified to use CRLF line endings.
+func toWindowsLineEndings(content []byte) []byte {
+	abnormal := false
+	for i, b := range content {
+		if b == '\n' && (i == 0 || content[i-1] != '\r') {
+			abnormal = true
+			break
+		}
+	}
+	if !abnormal {
+		return content
+	}
+	var buf bytes.Buffer
+	for i, b := range content {
+		if b == '\n' && (i == 0 || content[i-1] != '\r') {
+			buf.WriteByte('\r')
+		}
+		buf.WriteByte(b)
+	}
+	return buf.Bytes()
+}
+
 // CreateBuffer creates a new unsaved buffer corresponding to the workdir path,
 // containing the given textual content.
 func (e *Editor) CreateBuffer(ctx context.Context, path, content string) error {
-	return e.createBuffer(ctx, path, true, content)
+	return e.createBuffer(ctx, path, true, []byte(content))
 }
 
-func (e *Editor) createBuffer(ctx context.Context, path string, dirty bool, content string) error {
+func (e *Editor) createBuffer(ctx context.Context, path string, dirty bool, content []byte) error {
 	e.mu.Lock()
 
 	if _, ok := e.buffers[path]; ok {
@@ -397,12 +434,12 @@ func (e *Editor) createBuffer(ctx context.Context, path string, dirty bool, cont
 		return fmt.Errorf("buffer %q already exists", path)
 	}
 
+	uri := e.sandbox.Workdir.URI(path).SpanURI()
 	buf := buffer{
-		windowsLineEndings: e.config.WindowsLineEndings,
-		version:            1,
-		path:               path,
-		lines:              lines(content),
-		dirty:              dirty,
+		version: 1,
+		path:    path,
+		mapper:  protocol.NewMapper(uri, content),
+		dirty:   dirty,
 	}
 	e.buffers[path] = buf
 
@@ -464,15 +501,6 @@ func languageID(p string, fileAssociations map[string]string) string {
 	return ""
 }
 
-// lines returns line-ending agnostic line representation of content.
-func lines(content string) []string {
-	lines := strings.Split(content, "\n")
-	for i, l := range lines {
-		lines[i] = strings.TrimSuffix(l, "\r")
-	}
-	return lines
-}
-
 // CloseBuffer removes the current buffer (regardless of whether it is saved).
 func (e *Editor) CloseBuffer(ctx context.Context, path string) error {
 	e.mu.Lock()
@@ -566,48 +594,23 @@ func (e *Editor) SaveBufferWithoutActions(ctx context.Context, path string) erro
 	return nil
 }
 
-// contentPosition returns the (Line, Column) position corresponding to offset
-// in the buffer referenced by path.
-func contentPosition(content string, offset int) (Pos, error) {
-	scanner := bufio.NewScanner(strings.NewReader(content))
-	start := 0
-	line := 0
-	for scanner.Scan() {
-		end := start + len([]rune(scanner.Text())) + 1
-		if offset < end {
-			return Pos{Line: line, Column: offset - start}, nil
-		}
-		start = end
-		line++
-	}
-	if err := scanner.Err(); err != nil {
-		return Pos{}, fmt.Errorf("scanning content: %w", err)
-	}
-	// Scan() will drop the last line if it is empty. Correct for this.
-	if (strings.HasSuffix(content, "\n") || content == "") && offset == start {
-		return Pos{Line: line, Column: 0}, nil
-	}
-	return Pos{}, fmt.Errorf("position %d out of bounds in %q (line = %d, start = %d)", offset, content, line, start)
-}
-
 // ErrNoMatch is returned if a regexp search fails.
 var (
 	ErrNoMatch       = errors.New("no match")
 	ErrUnknownBuffer = errors.New("unknown buffer")
 )
 
-// regexpRange returns the start and end of the first occurrence of either re
+// regexpLocation returns the location of the first occurrence of either re
 // or its singular subgroup. It returns ErrNoMatch if the regexp doesn't match.
-func regexpRange(content, re string) (Pos, Pos, error) {
-	content = normalizeEOL(content)
+func regexpLocation(mapper *protocol.Mapper, re string) (protocol.Location, error) {
 	var start, end int
 	rec, err := regexp.Compile(re)
 	if err != nil {
-		return Pos{}, Pos{}, err
+		return protocol.Location{}, err
 	}
-	indexes := rec.FindStringSubmatchIndex(content)
+	indexes := rec.FindSubmatchIndex(mapper.Content)
 	if indexes == nil {
-		return Pos{}, Pos{}, ErrNoMatch
+		return protocol.Location{}, ErrNoMatch
 	}
 	switch len(indexes) {
 	case 2:
@@ -617,45 +620,26 @@ func regexpRange(content, re string) (Pos, Pos, error) {
 		// one subgroup: return its range
 		start, end = indexes[2], indexes[3]
 	default:
-		return Pos{}, Pos{}, fmt.Errorf("invalid search regexp %q: expect either 0 or 1 subgroups, got %d", re, len(indexes)/2-1)
+		return protocol.Location{}, fmt.Errorf("invalid search regexp %q: expect either 0 or 1 subgroups, got %d", re, len(indexes)/2-1)
 	}
-	startPos, err := contentPosition(content, start)
-	if err != nil {
-		return Pos{}, Pos{}, err
-	}
-	endPos, err := contentPosition(content, end)
-	if err != nil {
-		return Pos{}, Pos{}, err
-	}
-	return startPos, endPos, nil
+	return mapper.OffsetLocation(start, end)
 }
 
-func normalizeEOL(content string) string {
-	return strings.Join(lines(content), "\n")
-}
-
-// RegexpRange returns the first range in the buffer bufName matching re. See
-// RegexpSearch for more information on matching.
-func (e *Editor) RegexpRange(bufName, re string) (Pos, Pos, error) {
-	e.mu.Lock()
-	defer e.mu.Unlock()
-	buf, ok := e.buffers[bufName]
-	if !ok {
-		return Pos{}, Pos{}, ErrUnknownBuffer
-	}
-	return regexpRange(buf.text(), re)
-}
-
-// RegexpSearch returns the position of the first match for re in the buffer
+// RegexpSearch returns the Location of the first match for re in the buffer
 // bufName. For convenience, RegexpSearch supports the following two modes:
 //  1. If re has no subgroups, return the position of the match for re itself.
 //  2. If re has one subgroup, return the position of the first subgroup.
 //
 // It returns an error re is invalid, has more than one subgroup, or doesn't
 // match the buffer.
-func (e *Editor) RegexpSearch(bufName, re string) (Pos, error) {
-	start, _, err := e.RegexpRange(bufName, re)
-	return start, err
+func (e *Editor) RegexpSearch(bufName, re string) (protocol.Location, error) {
+	e.mu.Lock()
+	buf, ok := e.buffers[bufName]
+	e.mu.Unlock()
+	if !ok {
+		return protocol.Location{}, ErrUnknownBuffer
+	}
+	return regexpLocation(buf.mapper, re)
 }
 
 // RegexpReplace edits the buffer corresponding to path by replacing the first
@@ -670,20 +654,23 @@ func (e *Editor) RegexpReplace(ctx context.Context, path, re, replace string) er
 	if !ok {
 		return ErrUnknownBuffer
 	}
-	content := buf.text()
-	start, end, err := regexpRange(content, re)
+	loc, err := regexpLocation(buf.mapper, re)
 	if err != nil {
 		return err
 	}
-	return e.editBufferLocked(ctx, path, []Edit{{
-		Start: start,
-		End:   end,
-		Text:  replace,
-	}})
+	edits := []protocol.TextEdit{{
+		Range:   loc.Range,
+		NewText: replace,
+	}}
+	patched, err := applyEdits(buf.mapper, edits, e.config.WindowsLineEndings)
+	if err != nil {
+		return fmt.Errorf("editing %q: %v", path, err)
+	}
+	return e.setBufferContentLocked(ctx, path, true, patched, edits)
 }
 
 // EditBuffer applies the given test edits to the buffer identified by path.
-func (e *Editor) EditBuffer(ctx context.Context, path string, edits []Edit) error {
+func (e *Editor) EditBuffer(ctx context.Context, path string, edits []protocol.TextEdit) error {
 	e.mu.Lock()
 	defer e.mu.Unlock()
 	return e.editBufferLocked(ctx, path, edits)
@@ -692,8 +679,7 @@ func (e *Editor) EditBuffer(ctx context.Context, path string, edits []Edit) erro
 func (e *Editor) SetBufferContent(ctx context.Context, path, content string) error {
 	e.mu.Lock()
 	defer e.mu.Unlock()
-	lines := lines(content)
-	return e.setBufferContentLocked(ctx, path, true, lines, nil)
+	return e.setBufferContentLocked(ctx, path, true, []byte(content), nil)
 }
 
 // HasBuffer reports whether the file name is open in the editor.
@@ -704,11 +690,28 @@ func (e *Editor) HasBuffer(name string) bool {
 	return ok
 }
 
-// BufferText returns the content of the buffer with the given name.
-func (e *Editor) BufferText(name string) string {
+// BufferText returns the content of the buffer with the given name, or "" if
+// the file at that path is not open. The second return value reports whether
+// the file is open.
+func (e *Editor) BufferText(name string) (string, bool) {
 	e.mu.Lock()
 	defer e.mu.Unlock()
-	return e.buffers[name].text()
+	buf, ok := e.buffers[name]
+	if !ok {
+		return "", false
+	}
+	return buf.text(), true
+}
+
+// Mapper returns the protocol.Mapper for the given buffer name, if it is open.
+func (e *Editor) Mapper(name string) (*protocol.Mapper, error) {
+	e.mu.Lock()
+	defer e.mu.Unlock()
+	buf, ok := e.buffers[name]
+	if !ok {
+		return nil, fmt.Errorf("no mapper for %q", name)
+	}
+	return buf.mapper, nil
 }
 
 // BufferVersion returns the current version of the buffer corresponding to
@@ -719,24 +722,24 @@ func (e *Editor) BufferVersion(name string) int {
 	return e.buffers[name].version
 }
 
-func (e *Editor) editBufferLocked(ctx context.Context, path string, edits []Edit) error {
+func (e *Editor) editBufferLocked(ctx context.Context, path string, edits []protocol.TextEdit) error {
 	buf, ok := e.buffers[path]
 	if !ok {
 		return fmt.Errorf("unknown buffer %q", path)
 	}
-	content, err := applyEdits(buf.lines, edits)
+	content, err := applyEdits(buf.mapper, edits, e.config.WindowsLineEndings)
 	if err != nil {
 		return fmt.Errorf("editing %q: %v; edits:\n%v", path, err, edits)
 	}
 	return e.setBufferContentLocked(ctx, path, true, content, edits)
 }
 
-func (e *Editor) setBufferContentLocked(ctx context.Context, path string, dirty bool, content []string, fromEdits []Edit) error {
+func (e *Editor) setBufferContentLocked(ctx context.Context, path string, dirty bool, content []byte, fromEdits []protocol.TextEdit) error {
 	buf, ok := e.buffers[path]
 	if !ok {
 		return fmt.Errorf("unknown buffer %q", path)
 	}
-	buf.lines = content
+	buf.mapper = protocol.NewMapper(buf.mapper.URI, content)
 	buf.version++
 	buf.dirty = dirty
 	e.buffers[path] = buf
@@ -744,7 +747,7 @@ func (e *Editor) setBufferContentLocked(ctx context.Context, path string, dirty
 	// Otherwise, send the entire content.
 	var evts []protocol.TextDocumentContentChangeEvent
 	if len(fromEdits) == 1 {
-		evts = append(evts, fromEdits[0].toProtocolChangeEvent())
+		evts = append(evts, EditToChangeEvent(fromEdits[0]))
 	} else {
 		evts = append(evts, protocol.TextDocumentContentChangeEvent{
 			Text: buf.text(),
@@ -769,106 +772,86 @@ func (e *Editor) setBufferContentLocked(ctx context.Context, path string, dirty
 }
 
 // GoToDefinition jumps to the definition of the symbol at the given position
-// in an open buffer. It returns the path and position of the resulting jump.
-func (e *Editor) GoToDefinition(ctx context.Context, path string, pos Pos) (string, Pos, error) {
-	if err := e.checkBufferPosition(path, pos); err != nil {
-		return "", Pos{}, err
+// in an open buffer. It returns the location of the resulting jump.
+//
+// TODO(rfindley): rename to "Definition", to be consistent with LSP
+// terminology.
+func (e *Editor) GoToDefinition(ctx context.Context, loc protocol.Location) (protocol.Location, error) {
+	if err := e.checkBufferLocation(loc); err != nil {
+		return protocol.Location{}, err
 	}
 	params := &protocol.DefinitionParams{}
-	params.TextDocument.URI = e.sandbox.Workdir.URI(path)
-	params.Position = pos.ToProtocolPosition()
+	params.TextDocument.URI = loc.URI
+	params.Position = loc.Range.Start
 
 	resp, err := e.Server.Definition(ctx, params)
 	if err != nil {
-		return "", Pos{}, fmt.Errorf("definition: %w", err)
+		return protocol.Location{}, fmt.Errorf("definition: %w", err)
 	}
-	return e.extractFirstPathAndPos(ctx, resp)
+	return e.extractFirstLocation(ctx, resp)
 }
 
-// GoToTypeDefinition jumps to the type definition of the symbol at the given position
+// GoToTypeDefinition jumps to the type definition of the symbol at the given location
 // in an open buffer.
-func (e *Editor) GoToTypeDefinition(ctx context.Context, path string, pos Pos) (string, Pos, error) {
-	if err := e.checkBufferPosition(path, pos); err != nil {
-		return "", Pos{}, err
+func (e *Editor) GoToTypeDefinition(ctx context.Context, loc protocol.Location) (protocol.Location, error) {
+	if err := e.checkBufferLocation(loc); err != nil {
+		return protocol.Location{}, err
 	}
 	params := &protocol.TypeDefinitionParams{}
-	params.TextDocument.URI = e.sandbox.Workdir.URI(path)
-	params.Position = pos.ToProtocolPosition()
+	params.TextDocument.URI = loc.URI
+	params.Position = loc.Range.Start
 
 	resp, err := e.Server.TypeDefinition(ctx, params)
 	if err != nil {
-		return "", Pos{}, fmt.Errorf("type definition: %w", err)
+		return protocol.Location{}, fmt.Errorf("type definition: %w", err)
 	}
-	return e.extractFirstPathAndPos(ctx, resp)
+	return e.extractFirstLocation(ctx, resp)
 }
 
-// extractFirstPathAndPos returns the path and the position of the first location.
+// extractFirstLocation returns the first location.
 // It opens the file if needed.
-func (e *Editor) extractFirstPathAndPos(ctx context.Context, locs []protocol.Location) (string, Pos, error) {
+func (e *Editor) extractFirstLocation(ctx context.Context, locs []protocol.Location) (protocol.Location, error) {
 	if len(locs) == 0 {
-		return "", Pos{}, nil
+		return protocol.Location{}, nil
 	}
 
 	newPath := e.sandbox.Workdir.URIToPath(locs[0].URI)
-	newPos := fromProtocolPosition(locs[0].Range.Start)
 	if !e.HasBuffer(newPath) {
 		if err := e.OpenFile(ctx, newPath); err != nil {
-			return "", Pos{}, fmt.Errorf("OpenFile: %w", err)
+			return protocol.Location{}, fmt.Errorf("OpenFile: %w", err)
 		}
 	}
-	return newPath, newPos, nil
+	return locs[0], nil
 }
 
 // Symbol performs a workspace symbol search using query
-func (e *Editor) Symbol(ctx context.Context, query string) ([]SymbolInformation, error) {
-	params := &protocol.WorkspaceSymbolParams{}
-	params.Query = query
-
-	resp, err := e.Server.Symbol(ctx, params)
-	if err != nil {
-		return nil, fmt.Errorf("symbol: %w", err)
-	}
-	var res []SymbolInformation
-	for _, si := range resp {
-		ploc := si.Location
-		path := e.sandbox.Workdir.URIToPath(ploc.URI)
-		start := fromProtocolPosition(ploc.Range.Start)
-		end := fromProtocolPosition(ploc.Range.End)
-		rnge := Range{
-			Start: start,
-			End:   end,
-		}
-		loc := Location{
-			Path:  path,
-			Range: rnge,
-		}
-		res = append(res, SymbolInformation{
-			Name:     si.Name,
-			Kind:     si.Kind,
-			Location: loc,
-		})
-	}
-	return res, nil
+func (e *Editor) Symbol(ctx context.Context, query string) ([]protocol.SymbolInformation, error) {
+	params := &protocol.WorkspaceSymbolParams{Query: query}
+	return e.Server.Symbol(ctx, params)
 }
 
 // OrganizeImports requests and performs the source.organizeImports codeAction.
 func (e *Editor) OrganizeImports(ctx context.Context, path string) error {
-	_, err := e.applyCodeActions(ctx, path, nil, nil, protocol.SourceOrganizeImports)
+	loc := protocol.Location{URI: e.sandbox.Workdir.URI(path)} // zero Range => whole file
+	_, err := e.applyCodeActions(ctx, loc, nil, protocol.SourceOrganizeImports)
 	return err
 }
 
 // RefactorRewrite requests and performs the source.refactorRewrite codeAction.
-func (e *Editor) RefactorRewrite(ctx context.Context, path string, rng *protocol.Range) error {
-	applied, err := e.applyCodeActions(ctx, path, rng, nil, protocol.RefactorRewrite)
+func (e *Editor) RefactorRewrite(ctx context.Context, loc protocol.Location) error {
+	applied, err := e.applyCodeActions(ctx, loc, nil, protocol.RefactorRewrite)
+	if err != nil {
+		return err
+	}
 	if applied == 0 {
 		return fmt.Errorf("no refactorings were applied")
 	}
-	return err
+	return nil
 }
 
 // ApplyQuickFixes requests and performs the quickfix codeAction.
-func (e *Editor) ApplyQuickFixes(ctx context.Context, path string, rng *protocol.Range, diagnostics []protocol.Diagnostic) error {
-	applied, err := e.applyCodeActions(ctx, path, rng, diagnostics, protocol.SourceFixAll, protocol.QuickFix)
+func (e *Editor) ApplyQuickFixes(ctx context.Context, loc protocol.Location, diagnostics []protocol.Diagnostic) error {
+	applied, err := e.applyCodeActions(ctx, loc, diagnostics, protocol.SourceFixAll, protocol.QuickFix)
 	if applied == 0 {
 		return fmt.Errorf("no quick fixes were applied")
 	}
@@ -877,16 +860,17 @@ func (e *Editor) ApplyQuickFixes(ctx context.Context, path string, rng *protocol
 
 // ApplyCodeAction applies the given code action.
 func (e *Editor) ApplyCodeAction(ctx context.Context, action protocol.CodeAction) error {
-	for _, change := range action.Edit.DocumentChanges {
-		if change.TextDocumentEdit != nil {
-			path := e.sandbox.Workdir.URIToPath(change.TextDocumentEdit.TextDocument.URI)
-			if int32(e.buffers[path].version) != change.TextDocumentEdit.TextDocument.Version {
-				// Skip edits for old versions.
-				continue
-			}
-			edits := convertEdits(change.TextDocumentEdit.Edits)
-			if err := e.EditBuffer(ctx, path, edits); err != nil {
-				return fmt.Errorf("editing buffer %q: %w", path, err)
+	if action.Edit != nil {
+		for _, change := range action.Edit.DocumentChanges {
+			if change.TextDocumentEdit != nil {
+				path := e.sandbox.Workdir.URIToPath(change.TextDocumentEdit.TextDocument.URI)
+				if int32(e.buffers[path].version) != change.TextDocumentEdit.TextDocument.Version {
+					// Skip edits for old versions.
+					continue
+				}
+				if err := e.EditBuffer(ctx, path, change.TextDocumentEdit.Edits); err != nil {
+					return fmt.Errorf("editing buffer %q: %w", path, err)
+				}
 			}
 		}
 	}
@@ -905,12 +889,12 @@ func (e *Editor) ApplyCodeAction(ctx context.Context, action protocol.CodeAction
 }
 
 // GetQuickFixes returns the available quick fix code actions.
-func (e *Editor) GetQuickFixes(ctx context.Context, path string, rng *protocol.Range, diagnostics []protocol.Diagnostic) ([]protocol.CodeAction, error) {
-	return e.getCodeActions(ctx, path, rng, diagnostics, protocol.QuickFix, protocol.SourceFixAll)
+func (e *Editor) GetQuickFixes(ctx context.Context, loc protocol.Location, diagnostics []protocol.Diagnostic) ([]protocol.CodeAction, error) {
+	return e.getCodeActions(ctx, loc, diagnostics, protocol.QuickFix, protocol.SourceFixAll)
 }
 
-func (e *Editor) applyCodeActions(ctx context.Context, path string, rng *protocol.Range, diagnostics []protocol.Diagnostic, only ...protocol.CodeActionKind) (int, error) {
-	actions, err := e.getCodeActions(ctx, path, rng, diagnostics, only...)
+func (e *Editor) applyCodeActions(ctx context.Context, loc protocol.Location, diagnostics []protocol.Diagnostic, only ...protocol.CodeActionKind) (int, error) {
+	actions, err := e.getCodeActions(ctx, loc, diagnostics, only...)
 	if err != nil {
 		return 0, err
 	}
@@ -937,19 +921,17 @@ func (e *Editor) applyCodeActions(ctx context.Context, path string, rng *protoco
 	return applied, nil
 }
 
-func (e *Editor) getCodeActions(ctx context.Context, path string, rng *protocol.Range, diagnostics []protocol.Diagnostic, only ...protocol.CodeActionKind) ([]protocol.CodeAction, error) {
+func (e *Editor) getCodeActions(ctx context.Context, loc protocol.Location, diagnostics []protocol.Diagnostic, only ...protocol.CodeActionKind) ([]protocol.CodeAction, error) {
 	if e.Server == nil {
 		return nil, nil
 	}
 	params := &protocol.CodeActionParams{}
-	params.TextDocument.URI = e.sandbox.Workdir.URI(path)
+	params.TextDocument.URI = loc.URI
 	params.Context.Only = only
+	params.Range = loc.Range // may be zero => whole file
 	if diagnostics != nil {
 		params.Context.Diagnostics = diagnostics
 	}
-	if rng != nil {
-		params.Range = *rng
-	}
 	return e.Server.CodeAction(ctx, params)
 }
 
@@ -958,11 +940,13 @@ func (e *Editor) ExecuteCommand(ctx context.Context, params *protocol.ExecuteCom
 		return nil, nil
 	}
 	var match bool
-	// Ensure that this command was actually listed as a supported command.
-	for _, command := range e.serverCapabilities.ExecuteCommandProvider.Commands {
-		if command == params.Command {
-			match = true
-			break
+	if e.serverCapabilities.ExecuteCommandProvider != nil {
+		// Ensure that this command was actually listed as a supported command.
+		for _, command := range e.serverCapabilities.ExecuteCommandProvider.Commands {
+			if command == params.Command {
+				match = true
+				break
+			}
 		}
 	}
 	if !match {
@@ -975,19 +959,11 @@ func (e *Editor) ExecuteCommand(ctx context.Context, params *protocol.ExecuteCom
 	// Some commands use the go command, which writes directly to disk.
 	// For convenience, check for those changes.
 	if err := e.sandbox.Workdir.CheckForFileChanges(ctx); err != nil {
-		return nil, err
+		return nil, fmt.Errorf("checking for file changes: %v", err)
 	}
 	return result, nil
 }
 
-func convertEdits(protocolEdits []protocol.TextEdit) []Edit {
-	var edits []Edit
-	for _, lspEdit := range protocolEdits {
-		edits = append(edits, fromProtocolTextEdit(lspEdit))
-	}
-	return edits
-}
-
 // FormatBuffer gofmts a Go file.
 func (e *Editor) FormatBuffer(ctx context.Context, path string) error {
 	if e.Server == nil {
@@ -998,7 +974,7 @@ func (e *Editor) FormatBuffer(ctx context.Context, path string) error {
 	e.mu.Unlock()
 	params := &protocol.DocumentFormattingParams{}
 	params.TextDocument.URI = e.sandbox.Workdir.URI(path)
-	resp, err := e.Server.Formatting(ctx, params)
+	edits, err := e.Server.Formatting(ctx, params)
 	if err != nil {
 		return fmt.Errorf("textDocument/formatting: %w", err)
 	}
@@ -1007,24 +983,23 @@ func (e *Editor) FormatBuffer(ctx context.Context, path string) error {
 	if versionAfter := e.buffers[path].version; versionAfter != version {
 		return fmt.Errorf("before receipt of formatting edits, buffer version changed from %d to %d", version, versionAfter)
 	}
-	edits := convertEdits(resp)
 	if len(edits) == 0 {
 		return nil
 	}
 	return e.editBufferLocked(ctx, path, edits)
 }
 
-func (e *Editor) checkBufferPosition(path string, pos Pos) error {
+func (e *Editor) checkBufferLocation(loc protocol.Location) error {
 	e.mu.Lock()
 	defer e.mu.Unlock()
+	path := e.sandbox.Workdir.URIToPath(loc.URI)
 	buf, ok := e.buffers[path]
 	if !ok {
 		return fmt.Errorf("buffer %q is not open", path)
 	}
-	if !inText(pos, buf.lines) {
-		return fmt.Errorf("position %v is invalid in buffer %q", pos, path)
-	}
-	return nil
+
+	_, _, err := buf.mapper.RangeOffsets(loc.Range)
+	return err
 }
 
 // RunGenerate runs `go generate` non-recursively in the workdir-relative dir
@@ -1080,10 +1055,11 @@ func (e *Editor) CodeLens(ctx context.Context, path string) ([]protocol.CodeLens
 }
 
 // Completion executes a completion request on the server.
-func (e *Editor) Completion(ctx context.Context, path string, pos Pos) (*protocol.CompletionList, error) {
+func (e *Editor) Completion(ctx context.Context, loc protocol.Location) (*protocol.CompletionList, error) {
 	if e.Server == nil {
 		return nil, nil
 	}
+	path := e.sandbox.Workdir.URIToPath(loc.URI)
 	e.mu.Lock()
 	_, ok := e.buffers[path]
 	e.mu.Unlock()
@@ -1091,10 +1067,7 @@ func (e *Editor) Completion(ctx context.Context, path string, pos Pos) (*protoco
 		return nil, fmt.Errorf("buffer %q is not open", path)
 	}
 	params := &protocol.CompletionParams{
-		TextDocumentPositionParams: protocol.TextDocumentPositionParams{
-			TextDocument: e.TextDocumentIdentifier(path),
-			Position:     pos.ToProtocolPosition(),
-		},
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 	}
 	completions, err := e.Server.Completion(ctx, params)
 	if err != nil {
@@ -1105,19 +1078,20 @@ func (e *Editor) Completion(ctx context.Context, path string, pos Pos) (*protoco
 
 // AcceptCompletion accepts a completion for the given item at the given
 // position.
-func (e *Editor) AcceptCompletion(ctx context.Context, path string, pos Pos, item protocol.CompletionItem) error {
+func (e *Editor) AcceptCompletion(ctx context.Context, loc protocol.Location, item protocol.CompletionItem) error {
 	if e.Server == nil {
 		return nil
 	}
 	e.mu.Lock()
 	defer e.mu.Unlock()
+	path := e.sandbox.Workdir.URIToPath(loc.URI)
 	_, ok := e.buffers[path]
 	if !ok {
 		return fmt.Errorf("buffer %q is not open", path)
 	}
-	return e.editBufferLocked(ctx, path, convertEdits(append([]protocol.TextEdit{
+	return e.editBufferLocked(ctx, path, append([]protocol.TextEdit{
 		*item.TextEdit,
-	}, item.AdditionalTextEdits...)))
+	}, item.AdditionalTextEdits...))
 }
 
 // Symbols executes a workspace/symbols request on the server.
@@ -1151,12 +1125,13 @@ func (e *Editor) InlayHint(ctx context.Context, path string) ([]protocol.InlayHi
 	return hints, nil
 }
 
-// References returns references to the object at (path, pos), as returned by
+// References returns references to the object at loc, as returned by
 // the connected LSP server. If no server is connected, it returns (nil, nil).
-func (e *Editor) References(ctx context.Context, path string, pos Pos) ([]protocol.Location, error) {
+func (e *Editor) References(ctx context.Context, loc protocol.Location) ([]protocol.Location, error) {
 	if e.Server == nil {
 		return nil, nil
 	}
+	path := e.sandbox.Workdir.URIToPath(loc.URI)
 	e.mu.Lock()
 	_, ok := e.buffers[path]
 	e.mu.Unlock()
@@ -1164,10 +1139,7 @@ func (e *Editor) References(ctx context.Context, path string, pos Pos) ([]protoc
 		return nil, fmt.Errorf("buffer %q is not open", path)
 	}
 	params := &protocol.ReferenceParams{
-		TextDocumentPositionParams: protocol.TextDocumentPositionParams{
-			TextDocument: e.TextDocumentIdentifier(path),
-			Position:     pos.ToProtocolPosition(),
-		},
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 		Context: protocol.ReferenceContext{
 			IncludeDeclaration: true,
 		},
@@ -1179,24 +1151,25 @@ func (e *Editor) References(ctx context.Context, path string, pos Pos) ([]protoc
 	return locations, nil
 }
 
-// Rename performs a rename of the object at (path, pos) to newName, using the
+// Rename performs a rename of the object at loc to newName, using the
 // connected LSP server. If no server is connected, it returns nil.
-func (e *Editor) Rename(ctx context.Context, path string, pos Pos, newName string) error {
+func (e *Editor) Rename(ctx context.Context, loc protocol.Location, newName string) error {
 	if e.Server == nil {
 		return nil
 	}
+	path := e.sandbox.Workdir.URIToPath(loc.URI)
 
 	// Verify that PrepareRename succeeds.
 	prepareParams := &protocol.PrepareRenameParams{}
 	prepareParams.TextDocument = e.TextDocumentIdentifier(path)
-	prepareParams.Position = pos.ToProtocolPosition()
+	prepareParams.Position = loc.Range.Start
 	if _, err := e.Server.PrepareRename(ctx, prepareParams); err != nil {
 		return fmt.Errorf("preparing rename: %v", err)
 	}
 
 	params := &protocol.RenameParams{
 		TextDocument: e.TextDocumentIdentifier(path),
-		Position:     pos.ToProtocolPosition(),
+		Position:     loc.Range.Start,
 		NewName:      newName,
 	}
 	wsEdits, err := e.Server.Rename(ctx, params)
@@ -1211,13 +1184,14 @@ func (e *Editor) Rename(ctx context.Context, path string, pos Pos, newName strin
 	return nil
 }
 
-// Implementations returns implementations for the object at (path, pos), as
+// Implementations returns implementations for the object at loc, as
 // returned by the connected LSP server. If no server is connected, it returns
 // (nil, nil).
-func (e *Editor) Implementations(ctx context.Context, path string, pos Pos) ([]protocol.Location, error) {
+func (e *Editor) Implementations(ctx context.Context, loc protocol.Location) ([]protocol.Location, error) {
 	if e.Server == nil {
 		return nil, nil
 	}
+	path := e.sandbox.Workdir.URIToPath(loc.URI)
 	e.mu.Lock()
 	_, ok := e.buffers[path]
 	e.mu.Unlock()
@@ -1225,14 +1199,28 @@ func (e *Editor) Implementations(ctx context.Context, path string, pos Pos) ([]p
 		return nil, fmt.Errorf("buffer %q is not open", path)
 	}
 	params := &protocol.ImplementationParams{
-		TextDocumentPositionParams: protocol.TextDocumentPositionParams{
-			TextDocument: e.TextDocumentIdentifier(path),
-			Position:     pos.ToProtocolPosition(),
-		},
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 	}
 	return e.Server.Implementation(ctx, params)
 }
 
+func (e *Editor) SignatureHelp(ctx context.Context, loc protocol.Location) (*protocol.SignatureHelp, error) {
+	if e.Server == nil {
+		return nil, nil
+	}
+	path := e.sandbox.Workdir.URIToPath(loc.URI)
+	e.mu.Lock()
+	_, ok := e.buffers[path]
+	e.mu.Unlock()
+	if !ok {
+		return nil, fmt.Errorf("buffer %q is not open", path)
+	}
+	params := &protocol.SignatureHelpParams{
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
+	}
+	return e.Server.SignatureHelp(ctx, params)
+}
+
 func (e *Editor) RenameFile(ctx context.Context, oldPath, newPath string) error {
 	closed, opened, err := e.renameBuffers(ctx, oldPath, newPath)
 	if err != nil {
@@ -1331,8 +1319,7 @@ func (e *Editor) applyTextDocumentEdit(ctx context.Context, change protocol.Text
 			return err
 		}
 	}
-	fakeEdits := convertEdits(change.Edits)
-	return e.EditBuffer(ctx, path, fakeEdits)
+	return e.EditBuffer(ctx, path, change.Edits)
 }
 
 // Config returns the current editor configuration.
@@ -1408,10 +1395,12 @@ func (e *Editor) ChangeWorkspaceFolders(ctx context.Context, folders []string) e
 }
 
 // CodeAction executes a codeAction request on the server.
-func (e *Editor) CodeAction(ctx context.Context, path string, rng *protocol.Range, diagnostics []protocol.Diagnostic) ([]protocol.CodeAction, error) {
+// If loc.Range is zero, the whole file is implied.
+func (e *Editor) CodeAction(ctx context.Context, loc protocol.Location, diagnostics []protocol.Diagnostic) ([]protocol.CodeAction, error) {
 	if e.Server == nil {
 		return nil, nil
 	}
+	path := e.sandbox.Workdir.URIToPath(loc.URI)
 	e.mu.Lock()
 	_, ok := e.buffers[path]
 	e.mu.Unlock()
@@ -1423,9 +1412,7 @@ func (e *Editor) CodeAction(ctx context.Context, path string, rng *protocol.Rang
 		Context: protocol.CodeActionContext{
 			Diagnostics: diagnostics,
 		},
-	}
-	if rng != nil {
-		params.Range = *rng
+		Range: loc.Range, // may be zero
 	}
 	lens, err := e.Server.CodeAction(ctx, params)
 	if err != nil {
@@ -1435,22 +1422,22 @@ func (e *Editor) CodeAction(ctx context.Context, path string, rng *protocol.Rang
 }
 
 // Hover triggers a hover at the given position in an open buffer.
-func (e *Editor) Hover(ctx context.Context, path string, pos Pos) (*protocol.MarkupContent, Pos, error) {
-	if err := e.checkBufferPosition(path, pos); err != nil {
-		return nil, Pos{}, err
+func (e *Editor) Hover(ctx context.Context, loc protocol.Location) (*protocol.MarkupContent, protocol.Location, error) {
+	if err := e.checkBufferLocation(loc); err != nil {
+		return nil, protocol.Location{}, err
 	}
 	params := &protocol.HoverParams{}
-	params.TextDocument.URI = e.sandbox.Workdir.URI(path)
-	params.Position = pos.ToProtocolPosition()
+	params.TextDocument.URI = loc.URI
+	params.Position = loc.Range.Start
 
 	resp, err := e.Server.Hover(ctx, params)
 	if err != nil {
-		return nil, Pos{}, fmt.Errorf("hover: %w", err)
+		return nil, protocol.Location{}, fmt.Errorf("hover: %w", err)
 	}
 	if resp == nil {
-		return nil, Pos{}, nil
+		return nil, protocol.Location{}, nil
 	}
-	return &resp.Contents, fromProtocolPosition(resp.Range.Start), nil
+	return &resp.Contents, protocol.Location{URI: loc.URI, Range: resp.Range}, nil
 }
 
 func (e *Editor) DocumentLink(ctx context.Context, path string) ([]protocol.DocumentLink, error) {
@@ -1462,16 +1449,16 @@ func (e *Editor) DocumentLink(ctx context.Context, path string) ([]protocol.Docu
 	return e.Server.DocumentLink(ctx, params)
 }
 
-func (e *Editor) DocumentHighlight(ctx context.Context, path string, pos Pos) ([]protocol.DocumentHighlight, error) {
+func (e *Editor) DocumentHighlight(ctx context.Context, loc protocol.Location) ([]protocol.DocumentHighlight, error) {
 	if e.Server == nil {
 		return nil, nil
 	}
-	if err := e.checkBufferPosition(path, pos); err != nil {
+	if err := e.checkBufferLocation(loc); err != nil {
 		return nil, err
 	}
 	params := &protocol.DocumentHighlightParams{}
-	params.TextDocument.URI = e.sandbox.Workdir.URI(path)
-	params.Position = pos.ToProtocolPosition()
+	params.TextDocument.URI = loc.URI
+	params.Position = loc.Range.Start
 
 	return e.Server.DocumentHighlight(ctx, params)
 }
