@@ -5,9 +5,9 @@
 package lsp
 
 import (
+	"bytes"
 	"context"
 	"fmt"
-	"go/token"
 	"os"
 	"os/exec"
 	"path/filepath"
@@ -25,6 +25,7 @@ import (
 	"golang.org/x/tools/gopls/internal/lsp/tests/compare"
 	"golang.org/x/tools/gopls/internal/span"
 	"golang.org/x/tools/internal/bug"
+	"golang.org/x/tools/internal/diff"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/testenv"
 )
@@ -43,23 +44,17 @@ func TestMain(m *testing.M) {
 	os.Exit(m.Run())
 }
 
+// TestLSP runs the marker tests in files beneath testdata/ using
+// implementations of each of the marker operations (e.g. @codelens) that
+// make LSP RPCs (e.g. textDocument/codeLens) to a gopls server.
 func TestLSP(t *testing.T) {
 	tests.RunTests(t, "testdata", true, testLSP)
 }
 
-type runner struct {
-	server      *Server
-	data        *tests.Data
-	diagnostics map[span.URI][]*source.Diagnostic
-	ctx         context.Context
-	normalizers []tests.Normalizer
-	editRecv    chan map[span.URI]string
-}
-
 func testLSP(t *testing.T, datum *tests.Data) {
 	ctx := tests.Context(t)
 
-	session := cache.NewSession(ctx, cache.New(nil, nil), nil)
+	session := cache.NewSession(ctx, cache.New(nil), nil)
 	options := source.DefaultOptions().Clone()
 	tests.DefaultOptions(options)
 	session.SetOptions(options)
@@ -73,17 +68,41 @@ func testLSP(t *testing.T, datum *tests.Data) {
 
 	// Enable type error analyses for tests.
 	// TODO(golang/go#38212): Delete this once they are enabled by default.
-	tests.EnableAllAnalyzers(view, options)
+	tests.EnableAllAnalyzers(options)
 	session.SetViewOptions(ctx, view, options)
 
 	// Enable all inlay hints for tests.
-	tests.EnableAllInlayHints(view, options)
+	tests.EnableAllInlayHints(options)
 
 	// Only run the -modfile specific tests in module mode with Go 1.14 or above.
 	datum.ModfileFlagAvailable = len(snapshot.ModFiles()) > 0 && testenv.Go1Point() >= 14
 	release()
 
+	// Open all files for performance reasons. This is done because gopls only
+	// keeps active packages in memory for open files.
+	//
+	// In practice clients will only send document-oriented requests for open
+	// files.
 	var modifications []source.FileModification
+	for _, module := range datum.Exported.Modules {
+		for name := range module.Files {
+			filename := datum.Exported.File(module.Name, name)
+			if filepath.Ext(filename) != ".go" {
+				continue
+			}
+			content, err := datum.Exported.FileContents(filename)
+			if err != nil {
+				t.Fatal(err)
+			}
+			modifications = append(modifications, source.FileModification{
+				URI:        span.URIFromPath(filename),
+				Action:     source.Open,
+				Version:    -1,
+				Text:       content,
+				LanguageID: "go",
+			})
+		}
+	}
 	for filename, content := range datum.Config.Overlay {
 		if filepath.Ext(filename) != ".go" {
 			continue
@@ -103,13 +122,23 @@ func testLSP(t *testing.T, datum *tests.Data) {
 		data:        datum,
 		ctx:         ctx,
 		normalizers: tests.CollectNormalizers(datum.Exported),
-		editRecv:    make(chan map[span.URI]string, 1),
+		editRecv:    make(chan map[span.URI][]byte, 1),
 	}
 
 	r.server = NewServer(session, testClient{runner: r})
 	tests.Run(t, r, datum)
 }
 
+// runner implements tests.Tests by making LSP RPCs to a gopls server.
+type runner struct {
+	server      *Server
+	data        *tests.Data
+	diagnostics map[span.URI][]*source.Diagnostic
+	ctx         context.Context
+	normalizers []tests.Normalizer
+	editRecv    chan map[span.URI][]byte
+}
+
 // testClient stubs any client functions that may be called by LSP functions.
 type testClient struct {
 	protocol.Client
@@ -144,16 +173,13 @@ func (r *runner) CallHierarchy(t *testing.T, spn span.Span, expectedCalls *tests
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := mapper.Location(spn)
+	loc, err := mapper.SpanLocation(spn)
 	if err != nil {
 		t.Fatalf("failed for %v: %v", spn, err)
 	}
 
 	params := &protocol.CallHierarchyPrepareParams{
-		TextDocumentPositionParams: protocol.TextDocumentPositionParams{
-			TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-			Position:     loc.Range.Start,
-		},
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 	}
 
 	items, err := r.server.PrepareCallHierarchy(r.ctx, params)
@@ -220,8 +246,7 @@ func (r *runner) Diagnostics(t *testing.T, uri span.URI, want []*source.Diagnost
 	// Get the diagnostics for this view if we have not done it before.
 	v := r.server.session.View(r.data.Config.Dir)
 	r.collectDiagnostics(v)
-	got := append([]*source.Diagnostic(nil), r.diagnostics[uri]...) // copy
-	tests.CompareDiagnostics(t, uri, want, got)
+	tests.CompareDiagnostics(t, uri, want, r.diagnostics[uri])
 }
 
 func (r *runner) FoldingRanges(t *testing.T, spn span.Span) {
@@ -234,41 +259,30 @@ func (r *runner) FoldingRanges(t *testing.T, spn span.Span) {
 	modified := original
 	defer r.server.session.SetViewOptions(r.ctx, view, original)
 
-	// Test all folding ranges.
-	modified.LineFoldingOnly = false
-	view, err = r.server.session.SetViewOptions(r.ctx, view, modified)
-	if err != nil {
-		t.Error(err)
-		return
+	for _, test := range []struct {
+		lineFoldingOnly bool
+		prefix          string
+	}{
+		{false, "foldingRange"},
+		{true, "foldingRange-lineFolding"},
+	} {
+		modified.LineFoldingOnly = test.lineFoldingOnly
+		view, err = r.server.session.SetViewOptions(r.ctx, view, modified)
+		if err != nil {
+			t.Error(err)
+			continue
+		}
+		ranges, err := r.server.FoldingRange(r.ctx, &protocol.FoldingRangeParams{
+			TextDocument: protocol.TextDocumentIdentifier{
+				URI: protocol.URIFromSpanURI(uri),
+			},
+		})
+		if err != nil {
+			t.Error(err)
+			continue
+		}
+		r.foldingRanges(t, test.prefix, uri, ranges)
 	}
-	ranges, err := r.server.FoldingRange(r.ctx, &protocol.FoldingRangeParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(uri),
-		},
-	})
-	if err != nil {
-		t.Error(err)
-		return
-	}
-	r.foldingRanges(t, "foldingRange", uri, ranges)
-
-	// Test folding ranges with lineFoldingOnly = true.
-	modified.LineFoldingOnly = true
-	view, err = r.server.session.SetViewOptions(r.ctx, view, modified)
-	if err != nil {
-		t.Error(err)
-		return
-	}
-	ranges, err = r.server.FoldingRange(r.ctx, &protocol.FoldingRangeParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(uri),
-		},
-	})
-	if err != nil {
-		t.Error(err)
-		return
-	}
-	r.foldingRanges(t, "foldingRange-lineFolding", uri, ranges)
 }
 
 func (r *runner) foldingRanges(t *testing.T, prefix string, uri span.URI, ranges []protocol.FoldingRange) {
@@ -354,7 +368,7 @@ func conflict(a, b protocol.FoldingRange) bool {
 		(a.EndLine > b.StartLine || (a.EndLine == b.StartLine && a.EndCharacter > b.StartCharacter))
 }
 
-func foldRanges(m *protocol.ColumnMapper, contents string, ranges []protocol.FoldingRange) (string, error) {
+func foldRanges(m *protocol.Mapper, contents string, ranges []protocol.FoldingRange) (string, error) {
 	foldedText := "<>"
 	res := contents
 	// Apply the edits from the end of the file forward
@@ -362,11 +376,11 @@ func foldRanges(m *protocol.ColumnMapper, contents string, ranges []protocol.Fol
 	// TODO(adonovan): factor to use diff.ApplyEdits, which validates the input.
 	for i := len(ranges) - 1; i >= 0; i-- {
 		r := ranges[i]
-		start, err := m.Point(protocol.Position{r.StartLine, r.StartCharacter})
+		start, err := m.PositionPoint(protocol.Position{Line: r.StartLine, Character: r.StartCharacter})
 		if err != nil {
 			return "", err
 		}
-		end, err := m.Point(protocol.Position{r.EndLine, r.EndCharacter})
+		end, err := m.PositionPoint(protocol.Position{Line: r.EndLine, Character: r.EndCharacter})
 		if err != nil {
 			return "", err
 		}
@@ -378,11 +392,11 @@ func foldRanges(m *protocol.ColumnMapper, contents string, ranges []protocol.Fol
 func (r *runner) Format(t *testing.T, spn span.Span) {
 	uri := spn.URI()
 	filename := uri.Filename()
-	gofmted := string(r.data.Golden(t, "gofmt", filename, func() ([]byte, error) {
+	gofmted := r.data.Golden(t, "gofmt", filename, func() ([]byte, error) {
 		cmd := exec.Command("gofmt", filename)
 		out, _ := cmd.Output() // ignore error, sometimes we have intentionally ungofmt-able files
 		return out, nil
-	}))
+	})
 
 	edits, err := r.server.Formatting(r.ctx, &protocol.DocumentFormattingParams{
 		TextDocument: protocol.TextDocumentIdentifier{
@@ -390,7 +404,7 @@ func (r *runner) Format(t *testing.T, spn span.Span) {
 		},
 	})
 	if err != nil {
-		if gofmted != "" {
+		if len(gofmted) > 0 {
 			t.Error(err)
 		}
 		return
@@ -403,7 +417,7 @@ func (r *runner) Format(t *testing.T, spn span.Span) {
 	if err != nil {
 		t.Error(err)
 	}
-	if diff := compare.Text(gofmted, got); diff != "" {
+	if diff := compare.Bytes(gofmted, got); diff != "" {
 		t.Errorf("format failed for %s (-want +got):\n%s", filename, diff)
 	}
 }
@@ -442,6 +456,8 @@ func (r *runner) SemanticTokens(t *testing.T, spn span.Span) {
 }
 
 func (r *runner) Import(t *testing.T, spn span.Span) {
+	// Invokes textDocument/codeAction and applies all the "goimports" edits.
+
 	uri := spn.URI()
 	filename := uri.Filename()
 	actions, err := r.server.CodeAction(r.ctx, &protocol.CodeActionParams{
@@ -456,7 +472,7 @@ func (r *runner) Import(t *testing.T, spn span.Span) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	got := string(m.Content)
+	got := m.Content
 	if len(actions) > 0 {
 		res, err := applyTextDocumentEdits(r, actions[0].Edit.DocumentChanges)
 		if err != nil {
@@ -464,12 +480,11 @@ func (r *runner) Import(t *testing.T, spn span.Span) {
 		}
 		got = res[uri]
 	}
-	want := string(r.data.Golden(t, "goimports", filename, func() ([]byte, error) {
-		return []byte(got), nil
-	}))
-
-	if d := compare.Text(want, got); d != "" {
-		t.Errorf("import failed for %s:\n%s", filename, d)
+	want := r.data.Golden(t, "goimports", filename, func() ([]byte, error) {
+		return got, nil
+	})
+	if diff := compare.Bytes(want, got); diff != "" {
+		t.Errorf("import failed for %s:\n%s", filename, diff)
 	}
 }
 
@@ -484,7 +499,7 @@ func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []tests.S
 	if err != nil {
 		t.Fatal(err)
 	}
-	rng, err := m.Range(spn)
+	rng, err := m.SpanRange(spn)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -545,7 +560,7 @@ func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []tests.S
 	if !match {
 		t.Fatalf("unexpected kind for code action %s, got %v, want one of %v", action.Title, action.Kind, codeActionKinds)
 	}
-	var res map[span.URI]string
+	var res map[span.URI][]byte
 	if cmd := action.Command; cmd != nil {
 		_, err := r.server.ExecuteCommand(r.ctx, &protocol.ExecuteCommandParams{
 			Command:   action.Command.Command,
@@ -562,11 +577,11 @@ func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []tests.S
 		}
 	}
 	for u, got := range res {
-		want := string(r.data.Golden(t, "suggestedfix_"+tests.SpanName(spn), u.Filename(), func() ([]byte, error) {
-			return []byte(got), nil
-		}))
-		if want != got {
-			t.Errorf("suggested fixes failed for %s:\n%s", u.Filename(), compare.Text(want, got))
+		want := r.data.Golden(t, "suggestedfix_"+tests.SpanName(spn), u.Filename(), func() ([]byte, error) {
+			return got, nil
+		})
+		if diff := compare.Bytes(want, got); diff != "" {
+			t.Errorf("suggested fixes failed for %s:\n%s", u.Filename(), diff)
 		}
 	}
 }
@@ -578,7 +593,7 @@ func (r *runner) FunctionExtraction(t *testing.T, start span.Span, end span.Span
 		t.Fatal(err)
 	}
 	spn := span.New(start.URI(), start.Start(), end.End())
-	rng, err := m.Range(spn)
+	rng, err := m.SpanRange(spn)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -614,11 +629,11 @@ func (r *runner) FunctionExtraction(t *testing.T, start span.Span, end span.Span
 	}
 	res := <-r.editRecv
 	for u, got := range res {
-		want := string(r.data.Golden(t, "functionextraction_"+tests.SpanName(spn), u.Filename(), func() ([]byte, error) {
-			return []byte(got), nil
-		}))
-		if want != got {
-			t.Errorf("function extraction failed for %s:\n%s", u.Filename(), compare.Text(want, got))
+		want := r.data.Golden(t, "functionextraction_"+tests.SpanName(spn), u.Filename(), func() ([]byte, error) {
+			return got, nil
+		})
+		if diff := compare.Bytes(want, got); diff != "" {
+			t.Errorf("function extraction failed for %s:\n%s", u.Filename(), diff)
 		}
 	}
 }
@@ -630,7 +645,7 @@ func (r *runner) MethodExtraction(t *testing.T, start span.Span, end span.Span)
 		t.Fatal(err)
 	}
 	spn := span.New(start.URI(), start.Start(), end.End())
-	rng, err := m.Range(spn)
+	rng, err := m.SpanRange(spn)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -666,40 +681,39 @@ func (r *runner) MethodExtraction(t *testing.T, start span.Span, end span.Span)
 	}
 	res := <-r.editRecv
 	for u, got := range res {
-		want := string(r.data.Golden(t, "methodextraction_"+tests.SpanName(spn), u.Filename(), func() ([]byte, error) {
-			return []byte(got), nil
-		}))
-		if want != got {
-			t.Errorf("method extraction failed for %s:\n%s", u.Filename(), compare.Text(want, got))
+		want := r.data.Golden(t, "methodextraction_"+tests.SpanName(spn), u.Filename(), func() ([]byte, error) {
+			return got, nil
+		})
+		if diff := compare.Bytes(want, got); diff != "" {
+			t.Errorf("method extraction failed for %s:\n%s", u.Filename(), diff)
 		}
 	}
 }
 
-func (r *runner) Definition(t *testing.T, spn span.Span, d tests.Definition) {
+// TODO(rfindley): This handler needs more work. The output is still a bit hard
+// to read (range diffs do not format nicely), and it is too entangled with hover.
+func (r *runner) Definition(t *testing.T, _ span.Span, d tests.Definition) {
 	sm, err := r.data.Mapper(d.Src.URI())
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := sm.Location(d.Src)
+	loc, err := sm.SpanLocation(d.Src)
 	if err != nil {
 		t.Fatalf("failed for %v: %v", d.Src, err)
 	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-		Position:     loc.Range.Start,
-	}
-	var locs []protocol.Location
+	tdpp := protocol.LocationTextDocumentPositionParams(loc)
+	var got []protocol.Location
 	var hover *protocol.Hover
 	if d.IsType {
 		params := &protocol.TypeDefinitionParams{
 			TextDocumentPositionParams: tdpp,
 		}
-		locs, err = r.server.TypeDefinition(r.ctx, params)
+		got, err = r.server.TypeDefinition(r.ctx, params)
 	} else {
 		params := &protocol.DefinitionParams{
 			TextDocumentPositionParams: tdpp,
 		}
-		locs, err = r.server.Definition(r.ctx, params)
+		got, err = r.server.Definition(r.ctx, params)
 		if err != nil {
 			t.Fatalf("failed for %v: %+v", d.Src, err)
 		}
@@ -711,31 +725,41 @@ func (r *runner) Definition(t *testing.T, spn span.Span, d tests.Definition) {
 	if err != nil {
 		t.Fatalf("failed for %v: %v", d.Src, err)
 	}
-	if len(locs) != 1 {
-		t.Errorf("got %d locations for definition, expected 1", len(locs))
+	dm, err := r.data.Mapper(d.Def.URI())
+	if err != nil {
+		t.Fatal(err)
+	}
+	def, err := dm.SpanLocation(d.Def)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if !d.OnlyHover {
+		want := []protocol.Location{def}
+		if diff := cmp.Diff(want, got); diff != "" {
+			t.Fatalf("Definition(%s) mismatch (-want +got):\n%s", d.Src, diff)
+		}
 	}
 	didSomething := false
 	if hover != nil {
 		didSomething = true
 		tag := fmt.Sprintf("%s-hoverdef", d.Name)
-		expectHover := string(r.data.Golden(t, tag, d.Src.URI().Filename(), func() ([]byte, error) {
+		want := string(r.data.Golden(t, tag, d.Src.URI().Filename(), func() ([]byte, error) {
 			return []byte(hover.Contents.Value), nil
 		}))
-		got := tests.StripSubscripts(hover.Contents.Value)
-		expectHover = tests.StripSubscripts(expectHover)
-		if got != expectHover {
-			tests.CheckSameMarkdown(t, got, expectHover)
+		got := hover.Contents.Value
+		if diff := tests.DiffMarkdown(want, got); diff != "" {
+			t.Errorf("%s: markdown mismatch:\n%s", d.Src, diff)
 		}
 	}
 	if !d.OnlyHover {
 		didSomething = true
-		locURI := locs[0].URI.SpanURI()
+		locURI := got[0].URI.SpanURI()
 		lm, err := r.data.Mapper(locURI)
 		if err != nil {
 			t.Fatal(err)
 		}
-		if def, err := lm.Span(locs[0]); err != nil {
-			t.Fatalf("failed for %v: %v", locs[0], err)
+		if def, err := lm.LocationSpan(got[0]); err != nil {
+			t.Fatalf("failed for %v: %v", got[0], err)
 		} else if def != d.Def {
 			t.Errorf("for %v got %v want %v", d.Src, def, d.Def)
 		}
@@ -745,139 +769,88 @@ func (r *runner) Definition(t *testing.T, spn span.Span, d tests.Definition) {
 	}
 }
 
-func (r *runner) Implementation(t *testing.T, spn span.Span, impls []span.Span) {
+func (r *runner) Implementation(t *testing.T, spn span.Span, wantSpans []span.Span) {
 	sm, err := r.data.Mapper(spn.URI())
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := sm.Location(spn)
+	loc, err := sm.SpanLocation(spn)
 	if err != nil {
-		t.Fatalf("failed for %v: %v", spn, err)
+		t.Fatal(err)
 	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-		Position:     loc.Range.Start,
-	}
-	var locs []protocol.Location
-	params := &protocol.ImplementationParams{
-		TextDocumentPositionParams: tdpp,
-	}
-	locs, err = r.server.Implementation(r.ctx, params)
+	gotImpls, err := r.server.Implementation(r.ctx, &protocol.ImplementationParams{
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
+	})
 	if err != nil {
-		t.Fatalf("failed for %v: %v", spn, err)
+		t.Fatalf("Server.Implementation(%s): %v", spn, err)
 	}
-	if len(locs) != len(impls) {
-		t.Fatalf("got %d locations for implementation, expected %d", len(locs), len(impls))
+	gotLocs, err := tests.LocationsToSpans(r.data, gotImpls)
+	if err != nil {
+		t.Fatal(err)
 	}
-
-	var results []span.Span
-	for i := range locs {
-		locURI := locs[i].URI.SpanURI()
-		lm, err := r.data.Mapper(locURI)
-		if err != nil {
-			t.Fatal(err)
-		}
-		imp, err := lm.Span(locs[i])
-		if err != nil {
-			t.Fatalf("failed for %v: %v", locs[i], err)
-		}
-		results = append(results, imp)
+	sanitize := func(s string) string {
+		return strings.ReplaceAll(s, r.data.Config.Dir, "gopls/internal/lsp/testdata")
 	}
-	span.SortSpans(results) // to make tests
-	span.SortSpans(impls)   // deterministic
-
-	for i := range results {
-		if results[i] != impls[i] {
-			t.Errorf("for %dth implementation of %v got %v want %v", i, spn, results[i], impls[i])
-		}
+	want := sanitize(tests.SortAndFormatSpans(wantSpans))
+	got := sanitize(tests.SortAndFormatSpans(gotLocs))
+	if got != want {
+		t.Errorf("implementations(%s):\n%s", sanitize(fmt.Sprint(spn)), diff.Unified("want", "got", want, got))
 	}
 }
 
-func (r *runner) Highlight(t *testing.T, src span.Span, locations []span.Span) {
+func (r *runner) Highlight(t *testing.T, src span.Span, spans []span.Span) {
 	m, err := r.data.Mapper(src.URI())
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := m.Location(src)
+	loc, err := m.SpanLocation(src)
 	if err != nil {
-		t.Fatalf("failed for %v: %v", locations[0], err)
-	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-		Position:     loc.Range.Start,
+		t.Fatal(err)
 	}
 	params := &protocol.DocumentHighlightParams{
-		TextDocumentPositionParams: tdpp,
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 	}
 	highlights, err := r.server.DocumentHighlight(r.ctx, params)
 	if err != nil {
-		t.Fatal(err)
+		t.Fatalf("DocumentHighlight(%v) failed: %v", params, err)
 	}
-	if len(highlights) != len(locations) {
-		t.Fatalf("got %d highlights for highlight at %v:%v:%v, expected %d", len(highlights), src.URI().Filename(), src.Start().Line(), src.Start().Column(), len(locations))
+	var got []protocol.Range
+	for _, h := range highlights {
+		got = append(got, h.Range)
 	}
-	// Check to make sure highlights have a valid range.
-	var results []span.Span
-	for i := range highlights {
-		h, err := m.RangeSpan(highlights[i].Range)
-		if err != nil {
-			t.Fatalf("failed for %v: %v", highlights[i], err)
-		}
-		results = append(results, h)
-	}
-	// Sort results to make tests deterministic since DocumentHighlight uses a map.
-	span.SortSpans(results)
-	// Check to make sure all the expected highlights are found.
-	for i := range results {
-		if results[i] != locations[i] {
-			t.Errorf("want %v, got %v\n", locations[i], results[i])
-		}
-	}
-}
 
-func (r *runner) Hover(t *testing.T, src span.Span, text string) {
-	m, err := r.data.Mapper(src.URI())
-	if err != nil {
-		t.Fatal(err)
-	}
-	loc, err := m.Location(src)
-	if err != nil {
-		t.Fatalf("failed for %v", err)
-	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-		Position:     loc.Range.Start,
-	}
-	params := &protocol.HoverParams{
-		TextDocumentPositionParams: tdpp,
-	}
-	hover, err := r.server.Hover(r.ctx, params)
-	if err != nil {
-		t.Fatal(err)
-	}
-	if text == "" {
-		if hover != nil {
-			t.Errorf("want nil, got %v\n", hover)
-		}
-	} else {
-		if hover == nil {
-			t.Fatalf("want hover result to include %s, but got nil", text)
-		}
-		if got := hover.Contents.Value; got != text {
-			t.Errorf("want %v, got %v\n", text, got)
-		}
-		if want, got := loc.Range, hover.Range; want != got {
-			t.Errorf("want range %v, got %v instead", want, got)
+	var want []protocol.Range
+	for _, s := range spans {
+		rng, err := m.SpanRange(s)
+		if err != nil {
+			t.Fatalf("Mapper.SpanRange(%v) failed: %v", s, err)
 		}
+		want = append(want, rng)
+	}
+
+	sortRanges := func(s []protocol.Range) {
+		sort.Slice(s, func(i, j int) bool {
+			return protocol.CompareRange(s[i], s[j]) < 0
+		})
+	}
+
+	sortRanges(got)
+	sortRanges(want)
+
+	if diff := cmp.Diff(want, got); diff != "" {
+		t.Errorf("DocumentHighlight(%v) mismatch (-want +got):\n%s", src, diff)
 	}
 }
 
 func (r *runner) References(t *testing.T, src span.Span, itemList []span.Span) {
+	// This test is substantially the same as (*runner).References in source/source_test.go.
+	// TODO(adonovan): Factor (and remove fluff). Where should the common code live?
+
 	sm, err := r.data.Mapper(src.URI())
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := sm.Location(src)
+	loc, err := sm.SpanLocation(src)
 	if err != nil {
 		t.Fatalf("failed for %v: %v", src, err)
 	}
@@ -886,6 +859,8 @@ func (r *runner) References(t *testing.T, src span.Span, itemList []span.Span) {
 			want := make(map[protocol.Location]bool)
 			for i, pos := range itemList {
 				// We don't want the first result if we aren't including the declaration.
+				// TODO(adonovan): don't assume a single declaration:
+				// there may be >1 if corresponding methods are considered.
 				if i == 0 && !includeDeclaration {
 					continue
 				}
@@ -893,17 +868,14 @@ func (r *runner) References(t *testing.T, src span.Span, itemList []span.Span) {
 				if err != nil {
 					t.Fatal(err)
 				}
-				loc, err := m.Location(pos)
+				loc, err := m.SpanLocation(pos)
 				if err != nil {
 					t.Fatalf("failed for %v: %v", src, err)
 				}
 				want[loc] = true
 			}
 			params := &protocol.ReferenceParams{
-				TextDocumentPositionParams: protocol.TextDocumentPositionParams{
-					TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-					Position:     loc.Range.Start,
-				},
+				TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 				Context: protocol.ReferenceContext{
 					IncludeDeclaration: includeDeclaration,
 				},
@@ -912,13 +884,43 @@ func (r *runner) References(t *testing.T, src span.Span, itemList []span.Span) {
 			if err != nil {
 				t.Fatalf("failed for %v: %v", src, err)
 			}
-			if len(got) != len(want) {
-				t.Errorf("references failed: different lengths got %v want %v", len(got), len(want))
+
+			sanitize := func(s string) string {
+				// In practice, CONFIGDIR means "gopls/internal/lsp/testdata".
+				return strings.ReplaceAll(s, r.data.Config.Dir, "CONFIGDIR")
 			}
-			for _, loc := range got {
-				if !want[loc] {
-					t.Errorf("references failed: incorrect references got %v want %v", loc, want)
+			formatLocation := func(loc protocol.Location) string {
+				return fmt.Sprintf("%s:%d.%d-%d.%d",
+					sanitize(string(loc.URI)),
+					loc.Range.Start.Line+1,
+					loc.Range.Start.Character+1,
+					loc.Range.End.Line+1,
+					loc.Range.End.Character+1)
+			}
+			toSlice := func(set map[protocol.Location]bool) []protocol.Location {
+				// TODO(adonovan): use generic maps.Keys(), someday.
+				list := make([]protocol.Location, 0, len(set))
+				for key := range set {
+					list = append(list, key)
 				}
+				return list
+			}
+			toString := func(locs []protocol.Location) string {
+				// TODO(adonovan): use generic JoinValues(locs, formatLocation).
+				strs := make([]string, len(locs))
+				for i, loc := range locs {
+					strs[i] = formatLocation(loc)
+				}
+				sort.Strings(strs)
+				return strings.Join(strs, "\n")
+			}
+			gotStr := toString(got)
+			wantStr := toString(toSlice(want))
+			if gotStr != wantStr {
+				t.Errorf("incorrect references (got %d, want %d) at %s:\n%s",
+					len(got), len(want),
+					formatLocation(loc),
+					diff.Unified("want", "got", wantStr, gotStr))
 			}
 		})
 	}
@@ -949,7 +951,7 @@ func (r *runner) InlayHints(t *testing.T, spn span.Span) {
 			paddingRight = " "
 		}
 		edits[i] = protocol.TextEdit{
-			Range:   protocol.Range{Start: *hint.Position, End: *hint.Position},
+			Range:   protocol.Range{Start: hint.Position, End: hint.Position},
 			NewText: fmt.Sprintf("<%s%s%s>", paddingLeft, hint.Label[0].Value, paddingRight),
 		}
 	}
@@ -963,12 +965,12 @@ func (r *runner) InlayHints(t *testing.T, spn span.Span) {
 		t.Error(err)
 	}
 
-	withinlayHints := string(r.data.Golden(t, "inlayHint", filename, func() ([]byte, error) {
-		return []byte(got), nil
-	}))
+	withinlayHints := r.data.Golden(t, "inlayHint", filename, func() ([]byte, error) {
+		return got, nil
+	})
 
-	if withinlayHints != got {
-		t.Errorf("inlay hints failed for %s, expected:\n%v\ngot:\n%v", filename, withinlayHints, got)
+	if !bytes.Equal(withinlayHints, got) {
+		t.Errorf("inlay hints failed for %s, expected:\n%s\ngot:\n%s", filename, withinlayHints, got)
 	}
 }
 
@@ -981,24 +983,22 @@ func (r *runner) Rename(t *testing.T, spn span.Span, newText string) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := sm.Location(spn)
+	loc, err := sm.SpanLocation(spn)
 	if err != nil {
 		t.Fatalf("failed for %v: %v", spn, err)
 	}
 
 	wedit, err := r.server.Rename(r.ctx, &protocol.RenameParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(uri),
-		},
-		Position: loc.Range.Start,
-		NewName:  newText,
+		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
+		Position:     loc.Range.Start,
+		NewName:      newText,
 	})
 	if err != nil {
 		renamed := string(r.data.Golden(t, tag, filename, func() ([]byte, error) {
 			return []byte(err.Error()), nil
 		}))
 		if err.Error() != renamed {
-			t.Errorf("rename failed for %s, expected:\n%v\ngot:\n%v\n", newText, renamed, err)
+			t.Errorf("%s: rename failed for %s, expected:\n%v\ngot:\n%v\n", spn, newText, renamed, err)
 		}
 		return
 	}
@@ -1012,23 +1012,26 @@ func (r *runner) Rename(t *testing.T, spn span.Span, newText string) {
 	}
 	sort.Strings(orderedURIs)
 
-	var got string
+	// Print the name and content of each modified file,
+	// concatenated, and compare against the golden.
+	var buf bytes.Buffer
 	for i := 0; i < len(res); i++ {
 		if i != 0 {
-			got += "\n"
+			buf.WriteByte('\n')
 		}
 		uri := span.URIFromURI(orderedURIs[i])
 		if len(res) > 1 {
-			got += filepath.Base(uri.Filename()) + ":\n"
+			buf.WriteString(filepath.Base(uri.Filename()))
+			buf.WriteString(":\n")
 		}
-		val := res[uri]
-		got += val
+		buf.Write(res[uri])
 	}
-	want := string(r.data.Golden(t, tag, filename, func() ([]byte, error) {
-		return []byte(got), nil
-	}))
-	if want != got {
-		t.Errorf("rename failed for %s:\n%s", newText, compare.Text(want, got))
+	got := buf.Bytes()
+	want := r.data.Golden(t, tag, filename, func() ([]byte, error) {
+		return got, nil
+	})
+	if diff := compare.Bytes(want, got); diff != "" {
+		t.Errorf("rename failed for %s:\n%s", newText, diff)
 	}
 }
 
@@ -1037,16 +1040,12 @@ func (r *runner) PrepareRename(t *testing.T, src span.Span, want *source.Prepare
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := m.Location(src)
+	loc, err := m.SpanLocation(src)
 	if err != nil {
 		t.Fatalf("failed for %v: %v", src, err)
 	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-		Position:     loc.Range.Start,
-	}
 	params := &protocol.PrepareRenameParams{
-		TextDocumentPositionParams: tdpp,
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 	}
 	got, err := r.server.PrepareRename(context.Background(), params)
 	if err != nil {
@@ -1081,16 +1080,16 @@ func (r *runner) PrepareRename(t *testing.T, src span.Span, want *source.Prepare
 	}
 }
 
-func applyTextDocumentEdits(r *runner, edits []protocol.DocumentChanges) (map[span.URI]string, error) {
-	res := map[span.URI]string{}
+func applyTextDocumentEdits(r *runner, edits []protocol.DocumentChanges) (map[span.URI][]byte, error) {
+	res := make(map[span.URI][]byte)
 	for _, docEdits := range edits {
 		if docEdits.TextDocumentEdit != nil {
 			uri := docEdits.TextDocumentEdit.TextDocument.URI.SpanURI()
-			var m *protocol.ColumnMapper
+			var m *protocol.Mapper
 			// If we have already edited this file, we use the edited version (rather than the
 			// file in its original state) so that we preserve our initial changes.
 			if content, ok := res[uri]; ok {
-				m = protocol.NewColumnMapper(uri, []byte(content))
+				m = protocol.NewMapper(uri, content)
 			} else {
 				var err error
 				if m, err = r.data.Mapper(uri); err != nil {
@@ -1178,18 +1177,12 @@ func (r *runner) SignatureHelp(t *testing.T, spn span.Span, want *protocol.Signa
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := m.Location(spn)
+	loc, err := m.SpanLocation(spn)
 	if err != nil {
 		t.Fatalf("failed for %v: %v", loc, err)
 	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(spn.URI()),
-		},
-		Position: loc.Range.Start,
-	}
 	params := &protocol.SignatureHelpParams{
-		TextDocumentPositionParams: tdpp,
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 	}
 	got, err := r.server.SignatureHelp(r.ctx, params)
 	if err != nil {
@@ -1277,47 +1270,69 @@ func (r *runner) AddImport(t *testing.T, uri span.URI, expectedImport string) {
 	if want == nil {
 		t.Fatalf("golden file %q not found", uri.Filename())
 	}
-	if diff := compare.Text(got, string(want)); diff != "" {
+	if diff := compare.Bytes(want, got); diff != "" {
 		t.Errorf("%s mismatch\n%s", command.AddImport, diff)
 	}
 }
 
-func TestBytesOffset(t *testing.T) {
-	tests := []struct {
-		text string
-		pos  protocol.Position
-		want int
-	}{
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 0}, want: 0},
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 1}, want: 1},
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 2}, want: 1},
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 3}, want: 5},
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 4}, want: 6},
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 5}, want: -1},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 0, Character: 3}, want: 3},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 0, Character: 4}, want: 3},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 1, Character: 0}, want: 4},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 1, Character: 3}, want: 7},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 1, Character: 4}, want: 7},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 2, Character: 0}, want: 8},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 2, Character: 1}, want: -1},
-		{text: "aaa\nbbb\n\n", pos: protocol.Position{Line: 2, Character: 0}, want: 8},
+func (r *runner) SelectionRanges(t *testing.T, spn span.Span) {
+	uri := spn.URI()
+	sm, err := r.data.Mapper(uri)
+	if err != nil {
+		t.Fatal(err)
+	}
+	loc, err := sm.SpanLocation(spn)
+	if err != nil {
+		t.Error(err)
 	}
 
-	for i, test := range tests {
-		fname := fmt.Sprintf("test %d", i)
-		fset := token.NewFileSet()
-		f := fset.AddFile(fname, -1, len(test.text))
-		f.SetLinesForContent([]byte(test.text))
-		uri := span.URIFromPath(fname)
-		mapper := protocol.NewColumnMapper(uri, []byte(test.text))
-		got, err := mapper.Point(test.pos)
-		if err != nil && test.want != -1 {
-			t.Errorf("unexpected error: %v", err)
-		}
-		if err == nil && got.Offset() != test.want {
-			t.Errorf("want %d for %q(Line:%d,Character:%d), but got %d", test.want, test.text, int(test.pos.Line), int(test.pos.Character), got.Offset())
+	ranges, err := r.server.selectionRange(r.ctx, &protocol.SelectionRangeParams{
+		TextDocument: protocol.TextDocumentIdentifier{
+			URI: protocol.URIFromSpanURI(uri),
+		},
+		Positions: []protocol.Position{loc.Range.Start},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	sb := &strings.Builder{}
+	for i, path := range ranges {
+		fmt.Fprintf(sb, "Ranges %d: ", i)
+		rng := path
+		for {
+			s, e, err := sm.RangeOffsets(rng.Range)
+			if err != nil {
+				t.Error(err)
+			}
+
+			var snippet string
+			if e-s < 30 {
+				snippet = string(sm.Content[s:e])
+			} else {
+				snippet = string(sm.Content[s:s+15]) + "..." + string(sm.Content[e-15:e])
+			}
+
+			fmt.Fprintf(sb, "\n\t%v %q", rng.Range, strings.ReplaceAll(snippet, "\n", "\\n"))
+
+			if rng.Parent == nil {
+				break
+			}
+			rng = *rng.Parent
 		}
+		sb.WriteRune('\n')
+	}
+	got := sb.String()
+
+	testName := "selectionrange_" + tests.SpanName(spn)
+	want := r.data.Golden(t, testName, uri.Filename(), func() ([]byte, error) {
+		return []byte(got), nil
+	})
+	if want == nil {
+		t.Fatalf("golden file %q not found", uri.Filename())
+	}
+	if diff := compare.Text(got, string(want)); diff != "" {
+		t.Errorf("%s mismatch\n%s", testName, diff)
 	}
 }
 
@@ -1327,7 +1342,10 @@ func (r *runner) collectDiagnostics(view *cache.View) {
 	}
 	r.diagnostics = make(map[span.URI][]*source.Diagnostic)
 
-	snapshot, release := view.Snapshot(r.ctx)
+	snapshot, release, err := view.Snapshot()
+	if err != nil {
+		panic(err)
+	}
 	defer release()
 
 	// Always run diagnostics with analysis.
