@@ -33,7 +33,6 @@ func Format(ctx context.Context, snapshot Snapshot, fh FileHandle) ([]protocol.T
 		return nil, fmt.Errorf("can't format %q: file is generated", fh.URI().Filename())
 	}
 
-	fset := snapshot.FileSet()
 	pgf, err := snapshot.ParseGo(ctx, fh, ParseFull)
 	if err != nil {
 		return nil, err
@@ -54,6 +53,7 @@ func Format(ctx context.Context, snapshot Snapshot, fh FileHandle) ([]protocol.T
 	// This should be acceptable for all users, who likely be prompted to rebuild
 	// the LSP server on each Go release.
 	buf := &bytes.Buffer{}
+	fset := FileSetFor(pgf.Tok)
 	if err := format.Node(buf, fset, pgf.File); err != nil {
 		return nil, err
 	}
@@ -381,11 +381,11 @@ func FromProtocolEdits(m *protocol.Mapper, edits []protocol.TextEdit) ([]diff.Ed
 
 // ApplyProtocolEdits applies the patch (edits) to m.Content and returns the result.
 // It also returns the edits converted to diff-package form.
-func ApplyProtocolEdits(m *protocol.Mapper, edits []protocol.TextEdit) (string, []diff.Edit, error) {
+func ApplyProtocolEdits(m *protocol.Mapper, edits []protocol.TextEdit) ([]byte, []diff.Edit, error) {
 	diffEdits, err := FromProtocolEdits(m, edits)
 	if err != nil {
-		return "", nil, err
+		return nil, nil, err
 	}
-	out, err := diff.Apply(string(m.Content), diffEdits)
+	out, err := diff.ApplyBytes(m.Content, diffEdits)
 	return out, diffEdits, err
 }
