@@ -62,14 +62,6 @@ type Snapshot interface {
 	// on behalf of this snapshot.
 	BackgroundContext() context.Context
 
-	// Fileset returns the Fileset used to parse all the Go files in this snapshot.
-	//
-	// If the files are known to belong to a specific Package, use
-	// Package.FileSet instead. (We plan to eliminate the
-	// Snapshot's cache of parsed files, and thus the need for a
-	// snapshot-wide FileSet.)
-	FileSet() *token.FileSet
-
 	// ValidBuildConfiguration returns true if there is some error in the
 	// user's workspace. In particular, if they are both outside of a module
 	// and their GOPATH.
@@ -164,14 +156,6 @@ type Snapshot interface {
 	// excluding id itself.
 	ReverseDependencies(ctx context.Context, id PackageID, transitive bool) (map[PackageID]*Metadata, error)
 
-	// CachedImportPaths returns all the imported packages loaded in this
-	// snapshot, indexed by their package path (not import path, despite the name)
-	// and checked in TypecheckWorkspace mode.
-	//
-	// To reduce latency, it does not wait for type-checking to complete.
-	// It is intended for use only in completions.
-	CachedImportPaths(ctx context.Context) (map[PackagePath]*types.Package, error)
-
 	// ActiveMetadata returns a new, unordered slice containing
 	// metadata for all packages considered 'active' in the workspace.
 	//
@@ -183,7 +167,7 @@ type Snapshot interface {
 	AllMetadata(ctx context.Context) ([]*Metadata, error)
 
 	// Symbols returns all symbols in the snapshot.
-	Symbols(ctx context.Context) map[span.URI][]Symbol
+	Symbols(ctx context.Context) (map[span.URI][]Symbol, error)
 
 	// Metadata returns the metadata for the specified package,
 	// or nil if it was not found.
@@ -199,7 +183,28 @@ type Snapshot interface {
 
 	// TypeCheck parses and type-checks the specified packages,
 	// and returns them in the same order as the ids.
-	TypeCheck(ctx context.Context, mode TypecheckMode, ids ...PackageID) ([]Package, error)
+	// The resulting packages' types may belong to different importers,
+	// so types from different packages are incommensurable.
+	TypeCheck(ctx context.Context, ids ...PackageID) ([]Package, error)
+
+	// PackageDiagnostics returns diagnostics for files contained in specified
+	// packages.
+	//
+	// If these diagnostics cannot be loaded from cache, the requested packages
+	// may be type-checked.
+	PackageDiagnostics(ctx context.Context, ids ...PackageID) (map[span.URI][]*Diagnostic, error)
+
+	// References returns cross-references indexes for the specified packages.
+	//
+	// If these indexes cannot be loaded from cache, the requested packages may
+	// be type-checked.
+	References(ctx context.Context, ids ...PackageID) ([]XrefIndex, error)
+
+	// MethodSets returns method-set indexes for the specified packages.
+	//
+	// If these indexes cannot be loaded from cache, the requested packages may
+	// be type-checked.
+	MethodSets(ctx context.Context, ids ...PackageID) ([]*methodsets.Index, error)
 
 	// GetCriticalError returns any critical errors in the workspace.
 	//
@@ -207,6 +212,10 @@ type Snapshot interface {
 	GetCriticalError(ctx context.Context) *CriticalError
 }
 
+type XrefIndex interface {
+	Lookup(targets map[PackagePath]map[objectpath.Path]struct{}) (locs []protocol.Location)
+}
+
 // SnapshotLabels returns a new slice of labels that should be used for events
 // related to a snapshot.
 func SnapshotLabels(snapshot Snapshot) []label.Label {
@@ -220,7 +229,7 @@ func SnapshotLabels(snapshot Snapshot) []label.Label {
 //
 // Type-checking is expensive. Call snapshot.ParseGo if all you need
 // is a parse tree, or snapshot.MetadataForFile if you only need metadata.
-func PackageForFile(ctx context.Context, snapshot Snapshot, uri span.URI, mode TypecheckMode, pkgSel PackageSelector) (Package, *ParsedGoFile, error) {
+func PackageForFile(ctx context.Context, snapshot Snapshot, uri span.URI, pkgSel PackageSelector) (Package, *ParsedGoFile, error) {
 	metas, err := snapshot.MetadataForFile(ctx, uri)
 	if err != nil {
 		return nil, nil, err
@@ -234,7 +243,7 @@ func PackageForFile(ctx context.Context, snapshot Snapshot, uri span.URI, mode T
 	case WidestPackage:
 		metas = metas[len(metas)-1:]
 	}
-	pkgs, err := snapshot.TypeCheck(ctx, mode, metas[0].ID)
+	pkgs, err := snapshot.TypeCheck(ctx, metas[0].ID)
 	if err != nil {
 		return nil, nil, err
 	}
@@ -474,9 +483,12 @@ type Metadata struct {
 	DepsByPkgPath   map[PackagePath]PackageID // values are unique and non-empty
 	Module          *packages.Module
 	DepsErrors      []*packagesinternal.PackageError
-	LoadDir         string // directory from which go/packages was run
+	Diagnostics     []*Diagnostic // processed diagnostics from 'go list'
+	LoadDir         string        // directory from which go/packages was run
 }
 
+func (m *Metadata) String() string { return string(m.ID) }
+
 // IsIntermediateTestVariant reports whether the given package is an
 // intermediate test variant, e.g. "net/http [net/url.test]".
 //
@@ -591,33 +603,12 @@ const (
 	// This is the mode used when attempting to examine the package graph structure.
 	ParseHeader ParseMode = iota
 
-	// ParseExported specifies that the package is used only as a dependency,
-	// and only its exported declarations are needed. More may be included if
-	// necessary to avoid type errors.
-	ParseExported
-
 	// ParseFull specifies the full AST is needed.
 	// This is used for files of direct interest where the entire contents must
 	// be considered.
 	ParseFull
 )
 
-// AllParseModes contains all possible values of ParseMode.
-// It is used for cache invalidation on a file content change.
-var AllParseModes = []ParseMode{ParseHeader, ParseExported, ParseFull}
-
-// TypecheckMode controls what kind of parsing should be done (see ParseMode)
-// while type checking a package.
-type TypecheckMode int
-
-const (
-	// TypecheckFull means to use ParseFull.
-	TypecheckFull TypecheckMode = iota
-	// TypecheckWorkspace means to use ParseFull for workspace packages, and
-	// ParseExported for others.
-	TypecheckWorkspace
-)
-
 // A FileHandle is an interface to files tracked by the LSP session, which may
 // be either files read from disk, or open in the editor session (overlays).
 type FileHandle interface {
@@ -777,7 +768,15 @@ type (
 )
 
 // Package represents a Go package that has been parsed and type-checked.
-// It maintains only the relevant fields of a *go/packages.Package.
+//
+// By design, there is no way to reach from a Package to the Package
+// representing one of its dependencies.
+//
+// Callers must not assume that two Packages share the same
+// token.FileSet or types.Importer and thus have commensurable
+// token.Pos values or types.Objects. Instead, use stable naming
+// schemes, such as (URI, byte offset) for positions, or (PackagePath,
+// objectpath.Path) for exported declarations.
 type Package interface {
 	Metadata() *Metadata
 
@@ -792,10 +791,9 @@ type Package interface {
 	// Results of type checking:
 	GetTypes() *types.Package
 	GetTypesInfo() *types.Info
+	DependencyTypes(PackagePath) *types.Package // nil for indirect dependency of no consequence
 	HasTypeErrors() bool
 	DiagnosticsForFile(ctx context.Context, s Snapshot, uri span.URI) ([]*Diagnostic, error)
-	ReferencesTo(map[PackagePath]map[objectpath.Path]unit) []protocol.Location
-	MethodSetsIndex() *methodsets.Index
 }
 
 type unit = struct{}
@@ -827,7 +825,7 @@ type Diagnostic struct {
 	Message string
 
 	Tags    []protocol.DiagnosticTag
-	Related []RelatedInformation
+	Related []protocol.DiagnosticRelatedInformation
 
 	// Fields below are used internally to generate quick fixes. They aren't
 	// part of the LSP spec and don't leave the server.
