@@ -40,7 +40,7 @@ const (
 	Two
 )
 
-//go:generate stringer -type=Number
+//` + `go:generate stringer -type=Number
 `
 	tests := []struct {
 		label        string
@@ -78,6 +78,8 @@ const (
 // regression test for golang/go#39446. It also checks that these code lenses
 // only affect the diagnostics and contents of the containing go.mod file.
 func TestUpgradeCodelens(t *testing.T) {
+	testenv.NeedsGo1Point(t, 18) // uses go.work
+
 	const proxyWithLatest = `
 -- golang.org/x/hello@v1.3.3/go.mod --
 module golang.org/x/hello
@@ -185,11 +187,11 @@ require golang.org/x/hello v1.2.3
 				}); err != nil {
 					t.Fatal(err)
 				}
-				env.Await(env.DoneWithChangeWatchedFiles())
-				if got := env.Editor.BufferText("a/go.mod"); got != wantGoModA {
+				env.AfterChange()
+				if got := env.BufferText("a/go.mod"); got != wantGoModA {
 					t.Fatalf("a/go.mod upgrade failed:\n%s", compare.Text(wantGoModA, got))
 				}
-				if got := env.Editor.BufferText("b/go.mod"); got != wantGoModB {
+				if got := env.BufferText("b/go.mod"); got != wantGoModB {
 					t.Fatalf("b/go.mod changed unexpectedly:\n%s", compare.Text(wantGoModB, got))
 				}
 			})
@@ -201,35 +203,33 @@ require golang.org/x/hello v1.2.3
 				if vendoring {
 					env.RunGoCommandInDir("a", "mod", "vendor")
 				}
-				env.Await(env.DoneWithChangeWatchedFiles())
+				env.AfterChange()
 				env.OpenFile("a/go.mod")
 				env.OpenFile("b/go.mod")
 				env.ExecuteCodeLensCommand("a/go.mod", command.CheckUpgrades, nil)
 				d := &protocol.PublishDiagnosticsParams{}
-				env.Await(
-					OnceMet(
-						env.DiagnosticAtRegexpWithMessage("a/go.mod", `require`, "can be upgraded"),
-						ReadDiagnostics("a/go.mod", d),
-						// We do not want there to be a diagnostic for b/go.mod,
-						// but there may be some subtlety in timing here, where this
-						// should always succeed, but may not actually test the correct
-						// behavior.
-						env.NoDiagnosticAtRegexp("b/go.mod", `require`),
-					),
+				env.OnceMet(
+					Diagnostics(env.AtRegexp("a/go.mod", `require`), WithMessage("can be upgraded")),
+					ReadDiagnostics("a/go.mod", d),
+					// We do not want there to be a diagnostic for b/go.mod,
+					// but there may be some subtlety in timing here, where this
+					// should always succeed, but may not actually test the correct
+					// behavior.
+					NoDiagnostics(env.AtRegexp("b/go.mod", `require`)),
 				)
 				// Check for upgrades in b/go.mod and then clear them.
 				env.ExecuteCodeLensCommand("b/go.mod", command.CheckUpgrades, nil)
-				env.Await(env.DiagnosticAtRegexpWithMessage("b/go.mod", `require`, "can be upgraded"))
+				env.Await(Diagnostics(env.AtRegexp("b/go.mod", `require`), WithMessage("can be upgraded")))
 				env.ExecuteCodeLensCommand("b/go.mod", command.ResetGoModDiagnostics, nil)
-				env.Await(EmptyDiagnostics("b/go.mod"))
+				env.Await(NoDiagnostics(ForFile("b/go.mod")))
 
 				// Apply the diagnostics to a/go.mod.
 				env.ApplyQuickFixes("a/go.mod", d.Diagnostics)
-				env.Await(env.DoneWithChangeWatchedFiles())
-				if got := env.Editor.BufferText("a/go.mod"); got != wantGoModA {
+				env.AfterChange()
+				if got := env.BufferText("a/go.mod"); got != wantGoModA {
 					t.Fatalf("a/go.mod upgrade failed:\n%s", compare.Text(wantGoModA, got))
 				}
-				if got := env.Editor.BufferText("b/go.mod"); got != wantGoModB {
+				if got := env.BufferText("b/go.mod"); got != wantGoModB {
 					t.Fatalf("b/go.mod changed unexpectedly:\n%s", compare.Text(wantGoModB, got))
 				}
 			})
@@ -238,7 +238,6 @@ require golang.org/x/hello v1.2.3
 }
 
 func TestUnusedDependenciesCodelens(t *testing.T) {
-	testenv.NeedsGo1Point(t, 14)
 	const proxy = `
 -- golang.org/x/hello@v1.0.0/go.mod --
 module golang.org/x/hello
@@ -284,7 +283,7 @@ func main() {
 		env.OpenFile("go.mod")
 		env.ExecuteCodeLensCommand("go.mod", command.Tidy, nil)
 		env.Await(env.DoneWithChangeWatchedFiles())
-		got := env.Editor.BufferText("go.mod")
+		got := env.BufferText("go.mod")
 		const wantGoMod = `module mod.com
 
 go 1.14
@@ -299,8 +298,6 @@ require golang.org/x/hello v1.0.0
 
 func TestRegenerateCgo(t *testing.T) {
 	testenv.NeedsTool(t, "cgo")
-	testenv.NeedsGo1Point(t, 15)
-
 	const workspace = `
 -- go.mod --
 module example.com
@@ -321,18 +318,19 @@ func Foo() {
 	Run(t, workspace, func(t *testing.T, env *Env) {
 		// Open the file. We have a nonexistant symbol that will break cgo processing.
 		env.OpenFile("cgo.go")
-		env.Await(env.DiagnosticAtRegexpWithMessage("cgo.go", ``, "go list failed to return CompiledGoFiles"))
+		env.AfterChange(
+			Diagnostics(env.AtRegexp("cgo.go", ``), WithMessage("go list failed to return CompiledGoFiles")),
+		)
 
 		// Fix the C function name. We haven't regenerated cgo, so nothing should be fixed.
 		env.RegexpReplace("cgo.go", `int fortythree`, "int fortytwo")
 		env.SaveBuffer("cgo.go")
-		env.Await(OnceMet(
-			env.DoneWithSave(),
-			env.DiagnosticAtRegexpWithMessage("cgo.go", ``, "go list failed to return CompiledGoFiles"),
-		))
+		env.AfterChange(
+			Diagnostics(env.AtRegexp("cgo.go", ``), WithMessage("go list failed to return CompiledGoFiles")),
+		)
 
 		// Regenerate cgo, fixing the diagnostic.
 		env.ExecuteCodeLensCommand("cgo.go", command.RegenerateCgo, nil)
-		env.Await(EmptyDiagnostics("cgo.go"))
+		env.Await(NoDiagnostics(ForFile("cgo.go")))
 	})
 }
